/**
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  09/29/00:  Created.
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;


/**
 * ArrowButton is custom Button widget. It works just like awt Button.
 * <BR><BR>
 * ArrowButton has four directions defined in CConstants class. The 
 * default direction is DIR_UP. It can be set using setDirection method.
 * 
 *
 * @version   1.0  09/29/00
 * @author    Ying Chen, Chip Watson,
 * @since     adm1.0
 *
 */
public class ArrowButton extends Canvas {  
   
    static final int WIDTH = 10;
    static final int HEIGHT = 10;

    //--- used for double-buffering ------------------------------
    protected Dimension offDimension;     // double-buffer size
    protected Image     offImage;         // off-screen buffer
    protected Graphics  offGraphics;      // off-screen paint handle

    transient ActionListener actionListener;
    private int direction = CConstants.DIR_UP;
    private Polygon triangle = new Polygon();
    private boolean pressed = false;

    /**
     * Constructor (dufault direction is DIR_UP)
     */
    public ArrowButton()
    {
	this(CConstants.DIR_UP);
    }


    /**
     * Constructor with specific direction
     *
     * @param dir  the button arrow direction, which may be one of:
     *              <OL>
     *                <LI>  CConstants.DIR_UP
     *                <LI>  CConstants.DIR_DOWN
     *                <LI>  CConstants.DIR_RIGHT
     *                <LI>  CConstants.DIR_LEFTL
     *              </OL>
     */
    public ArrowButton (int dir)  {
	setSize(WIDTH, HEIGHT);
	setDirection(dir);
	enableEvents(AWTEvent.MOUSE_EVENT_MASK);
	repaint();
    }

    /** 
     * set the direction of the button arrow
     * @param dir  the arrow direction, which may be one of:
     *              <OL>
     *                <LI>  CConstants.DIR_UP
     *                <LI>  CConstants.DIR_DOWN
     *                <LI>  CConstants.DIR_RIGHT
     *                <LI>  CConstants.DIR_LEFTL
     *              </OL>
     */
    public void setDirection (int dir) {
	direction = dir;
	repaint();
    }

    /**
     * Return the direction of the arrow button
     */
    public int getDirection () {
	return direction;
    }

    /**
     * Adds the specified action listener to receive action events from
     * this button. Action events occur when a user presses or releases
     * the mouse over this button.
     * @param         l the action listener.
     */ 
    public synchronized void addActionListener(ActionListener l) {
        actionListener = AWTEventMulticaster.add(actionListener, l);
    }

    /**
     * Removes the specified action listener so that it no longer 
     * receives action events from this button. Action events occur  
     * when a user presses or releases the mouse over this button.
     * @param         l     the action listener.
     */ 
    public synchronized void removeActionListener(ActionListener l) {
        actionListener = AWTEventMulticaster.remove(actionListener, l);
    }

    protected void processActionEvent(ActionEvent e) {
        if (actionListener != null) {
            actionListener.actionPerformed(e);
        }
    }
    
    /**
     * Draw the arrow button
     */
    public final void paint (Graphics g) {
	if (isShowing()) {
	    if (offGraphics == null) 
		setupOffscreen();
	}
	drawArrow(offGraphics); 
	g.drawImage(offImage, 0, 0, null);
    }

   /**
     * Set up offscreen graphics.  This is needed for both 
     * a new screen paint and a resize paint, or anything which
     * changes the widget's background (color, font, etc.).
     */
    protected void setupOffscreen() {
        offDimension = getSize();
        offImage = createImage(offDimension.width, offDimension.height);
        offGraphics = offImage.getGraphics();
        offGraphics.setColor(getBackground());
        offGraphics.fillRect(0, 0, offDimension.width, offDimension.height);
    }


    private void drawArrow (Graphics g) {
	Dimension rec = getSize();
	int w = rec.width;
	int h = rec.height;
	int x1, y1, x2, y2, x3, y3;
	switch(direction) {
	case CConstants.DIR_RIGHT:
	    x1 = 0;  y1 = 0;
	    x2 = 0;  y2 = h;
	    x3 = w;  y3 = (int) h/2;
	    break;
	case CConstants.DIR_LEFT:
	    x1 = w;  y1 = 0;
	    x2 = 0;  y2 = (int) h/2;
	    x3 = w;  y3 = h;
	    break;
	case CConstants.DIR_DOWN:
	    x1 = 0;          y1 = 0;
	    x2 = (int) w/2;  y2 = h;
	    x3 = w;          y3 = 0;
	    break;
	default:        //case DIR_UP:
	    x1 = (int) w/2;  y1 = 0;
	    x2 = 0;          y2 = h;
	    x3 = w;          y3 = h;
	}
	triangle.addPoint (x1, y1);
	triangle.addPoint (x3, y3);
	triangle.addPoint (x2, y2);
	g.setColor(getForeground());
        g.fillPolygon (triangle); 
    }

    /**
     * Handle the mouse event. Release the mouse to generate the 
     * ActionEvent
     */
    public void processMouseEvent (MouseEvent e) 
    {
	if (e.getID() == MouseEvent.MOUSE_PRESSED) 
	    pressed = true;
	else if (e.getID() == MouseEvent.MOUSE_RELEASED) {
	    //if (pressed && triangle.contains(e.getPoint())) {
	    if (pressed) {
		ActionEvent ae = 
		new ActionEvent(this, ActionEvent.ACTION_PERFORMED, "");
		processActionEvent(ae);
		pressed = false;
	    }
	}
    }

}
	


