/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  03/10/00:  Release version 0.1.
 *  04/20/00:  Release version 0.2.  
 *
 */

package adm;

import java.awt.*;
import cdev.*;

/** 
 * CBar is an awt component bean. It displays a property's value in the 
 * control system as a moving rectangular filled area. The size, bound, 
 * foreground, background and the font of the label on the bar can be set 
 * like as for an awt component. 
 * <BR><BR>
 * The color of the filled rectangular is the same as the foreground if the 
 * color mode property is set to CLRMOD_STATIC (default). It will change 
 * from foreground color to yellow or red depending on the property's value 
 * and warning and alarm limits when CLRMOD_ALARM is used. 
 * <BR><BR>
 * There are four label styles; the default label style is LABEL_NONE. If
 * LABEL_AXIS is used; the tick marks and display range will appear on the 
 * CBar widget. In addition to the tick marks and display range, the 
 * property's value will also shown on the widget if LABEL_VALUE is selected. 
 * All the information including the device and property name will on the 
 * widget when LABEL_ALL is used. 
 * <BR><BR>
 * The orientation of the CBar can be either VERTICAL (default) or HORIZONTAL. 
 * Finally the fill style can be set to FILL_FROM_EDGE (default) or
 * FILL_FROM_CENTER using the setFillStyle (int) method. 
 *
 *
 * @version   0.2  04/17/00
 * @author    Ying Chen, Chip Watson
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.1
 *
 */
public class CBar extends CIndicator {
 
    static final int LOWESTLOW = 2;      	// smallest bar fill.
    private int fillMode = FILL_FROM_EDGE;
    
    
    /**
     * Null Constructor.
     */
    public CBar() 
    {
	super();	
    }
    
    
    /**
     * Constructor with device name and property name.
     * 
     * @param dName    device name
     * @param pName    property name
     */
    public CBar(String dName, String pName)
    {
	super(dName, pName);	
    }
    
    
    /**
     * Constructor
     * Defines the canvas size and use default foreground and background color
     * @param width    component width
     * @param height   component height
     */
    public CBar (int w, int h) 
    {
	super(w, h);	
    }
    
    
    /**
     * Constructor
     * Defines the canvas size and sets foreground and background colors.
     * @param width     component width
     * @param height    component height
     * @param fg        foreground color
     * @param bg        background color
     */
    public CBar(int w, int h, Color fg, Color bg) 
    {
	super(w, h, fg, bg);	
    }
    
    /**
     * Get the fill mode
     */
    public int getFillMode () { return fillMode; }
    
    
    /**
     * Set the fill mode
     * @param newFillMode   new fill mode, which may be one of:
     *             <OL>
     *             <LI>  CConstants.FILL_FROM_EDGE
     *             <LI>  CConstants.FILL_FROM_CENTER
     *             </OL>
     */
    public void setFillMode (int newFillMode)
    {
	fillMode = newFillMode;
	offGraphics = null;
    }
    
    
    protected void paintForeground (Graphics g) 
    {
	if ((labelStyle == LABEL_VALUE)||(labelStyle == LABEL_ALL)) {
	    drawValue(g);
	}
	
	//--------------------------------------------
	// Erase old fill rect.
	//--------------------------------------------
	g.setColor (bgColor);
	if (orientation == VERTICAL)
	    g.fill3DRect(fillX, faceY+frameW, fillW, maxFillH, false);
	else
	    g.fill3DRect(faceX+frameW, fillY, maxFillW, fillH, false);
	
	if (rangeStyle == RANGE_AUTO)   do_auto();
	
	//------------------------------------------------------
	// recalculate fill% 0.5<=fill%>=-0.5(fill from center)
	//------------------------------------------------------
	if (fillMode == FILL_FROM_CENTER) {
	    fillPercent = (currVal-(rangeHi+rangeLo)/2)/(rangeHi-rangeLo);
	    fillPercent = Math.min(fillPercent, 0.5);      
	    fillPercent = Math.max(fillPercent, -0.5); 
	    if (orientation == VERTICAL) {
		fillH = (int)(maxFillH*Math.abs(fillPercent));
		fillH = Math.max(fillH, LOWESTLOW); 
		fillY = faceY+frameW+(int)(maxFillH/2)-(int)(maxFillH*(Math.max(fillPercent,0.0)));
	    }
	    else {
		fillW = (int) (maxFillW*Math.abs(fillPercent));
		fillW = Math.max(fillW, LOWESTLOW);
		fillX = faceX+frameW+(int)(maxFillW/2)+(int)(maxFillW*(Math.min(fillPercent, 0.0)));
	    }
	}
	else {  // Fill from edge
	    //------------------------------------------------------
	    // recalculate fill% 1.0<=fill%>=0.0 (fill from edge)
	    //------------------------------------------------------ 
	    fillPercent = (currVal-rangeLo)/(rangeHi-rangeLo);
	    fillPercent = Math.min(fillPercent, 1.0);   
	    fillPercent = Math.max(fillPercent, 0.0);	
	    if (orientation == HORIZONTAL ) {
		fillW = (int)(maxFillW*fillPercent);
		fillW = Math.max(fillW, LOWESTLOW);
	    }
	    else {           //(orientation == VERTICAL) 
		fillH = (int)(maxFillH*fillPercent); 
		fillH = Math.max(fillH, LOWESTLOW);
		fillY = faceY + faceH - frameW - fillH;
	    }
	}	
	
	//--------------------------------------------
	// Redraw new fill rect.
	//--------------------------------------------
	g.setColor(currColor);
	if (orientation == HORIZONTAL ) {
	    g.fillRect (fillX, fillY, fillW, fillH-1);
	}
	else {
	    g.fillRect (fillX, fillY, fillW-1, fillH);
	}
    }
    
} //end of CBar









