/**
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  04/17/00:  Release version 0.2.
 *  06/22/00:  Release version 0.3 (add controllable function)
 *  09/06/00:  Release version 0.4 (add isControllable() method)
 *  11/07/00:  Release version 1.0 (remove deprecated getControllable())
 *
 */

package adm;

import java.io.*; 
import java.awt.*;
import java.awt.event.*;
import cdev.*;
import cdev.data.*;


/**
 * CBits is an awt component bean. It displays and controls an integer 
 * property's value in the control system as bit patterns. The size, bound, 
 * foreground, background and the font of the label on the widget can be 
 * set like an awt component. 
 * <BR><BR>
 * The bits display range can be set through the setDisplayBits (int, int), 
 * setStartBit (int) and setEndBit (int) methods. The default bits display 
 * range is 0 to 15. 
 * <BR><BR>
 * By default it is a monitor widget only (the controllable property is set 
 * to false). Since version 0.3 it is can used to control the individual 
 * bits by calling setControllable(true). If the controllable property is 
 * set to true, the individual bit rectangle works like a push button. 
 * Pushing the bit button the value will change from 0 to 1 or 1 to 0. 
 * <BR><BR>
 * There are three label styles with a default of LABEL_NONE. If LABEL_AXIS 
 * (or LABEL_VALUE) is used, start/end bit numbers and a arrow indicating 
 * increasing bit number will be shown. All the information including the 
 * device and property name is shown when LABEL_ALL is used. 
 * <BR><BR>
 * CBits has four directions which can be selected. The bits can start from 
 * right (direction = DIR_LEFT), left (direction = DIR_RIGHT), top (direction 
 * = DIR_DOWN) or bottom (direction = DIR_UP). 
 *
 *
 * @version   0.2  03/29/00
 * @author    Ying Chen, Chip Watson, 
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.2
 *
 */
public class CBits extends CAnimatedCanvas implements MouseListener {

    private static int VAL_TAG = Data.convertKey("value");

    //--- GUI graphics info --------------------------------------
    private int cellCount;
    private int fillX, fillY, fillW, fillH = 0;  
    private int axisX, axisY, axisW, axisH = 0;
    private int cellX, cellY, cellW, cellH = 0;   
    private int chaX, chaY, chaW, chaH = 0;	
    private FontMetrics fm;
    private int fontH, lowLen, highLen, maxLabel = 0;
    private String stringHi, stringLo;

    //--- define default value-------- --------------------------
    private boolean firstCall = true;
    private boolean controllable = false;
    private int direction = DIR_LEFT;
    private int startBit = 0;
    private int endBit = 15;
    private int value = 0;
    private boolean[] bitVal;
    private Data tdata = new Data();

    /**
     * Null constructor. 
     */
    public CBits() { }

    /**
     * Constructor with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CBits(String dName, String pName)
    {
        super(dName, pName);
    }

    /**
     * Construct a CBits class with specific direction. 
     * @param direction    the direction of the bits display, which be one of:
     *                     <OL>
     *                     <LI>  CConstants.DIR_UP
     *                     <LI>  CConstants.DIR_DOWN
     *                     <LI>  CConstants.DIR_RIGHT
     *                     <LI>  CConstants.DIR_LEFT
     *                     </OL>
     */
    public CBits(int direction) {
	this.direction = direction; 
    }

    /**
     *  Return the startBit
     */
    public int getStartBit () { return startBit; }

    /**
     *  Return the endBit
     */
    public int getEndBit () { return endBit; }


    /**
     * Set the startBit. sBit must lesser than endBit
     * otherwise nothing is changed.
     * @param sBit    start bit position
     */
    public void setStartBit (int sBit)
    {
        if (sBit < endBit) {
            startBit = sBit;
            offGraphics = null;
        }
    }

    /**
     * Set the endBit. eBit must larger or equal than startBit
     * otherwise nothing is changed.
     * @param eBit    end bit position
     */
    public void setEndBit (int eBit)
    {
        if (eBit >= startBit) {
            endBit = eBit;
            offGraphics = null;
        }
    }

    /**
     * Set the startBit and endBit. eBit must larger or equal than sBit
     * otherwise nothing is changed.
     * @param sBit    start bit position
     * @param eBit    end bit position
     */
    public void setDisplayBits (int sBit, int eBit)
    {
        if (eBit >= sBit) {
            startBit = sBit;
            endBit = eBit;
            offGraphics = null;
        }
    }

    /**
     * Return the direction
     */
    public int getDirection () { return direction; }

    /**
     * Set bit position direction
     * @param newDir    the new direction, which be one of:
     *                 <OL>
     *                 <LI>  CConstants.DIR_UP
     *                 <LI>  CConstan.DIR_DOWN
     *                 <LI>  CConstants.DIR_RIGHT
     *                 <LI>  CConstan.DIR_LEFT
     *                 </OL>
     */
    public void setDirection(int newDir) 
    {
	direction = newDir;
	offGraphics = null;
    }
 
    /**
     * Set the controllable parameter.
     * @param b    = true (this widget can used to control value)
     *             = false (only used to monitor the value)
     */
    public void setControllable (boolean b) { 
	controllable = b; 
	if (controllable)  {
	    addMouseListener(this);
	}
	else {
	    removeMouseListener(this);
	}
    }

    /**
     * Return true if this widget is controllabel.
     * Rerurn flase if not
     */
    public boolean isControllable () { return controllable; }


    private void initGraphics () 
    {   
        Dimension rec = getSize ();
	chaX = 0; chaY = 0; chaW = 0; chaH = 0;
        axisX= 0; axisY= 0; axisW= 0; axisH= 0;

       	cellCount = Math.abs(endBit - startBit) + 1;
	bitVal = new boolean[endBit+1];
	
	if (labelStyle != LABEL_NONE) {
	    //fm = getFontMetrics(font); // for java 1.2
	    fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
	    fontH = fm.getHeight();
	    stringLo = Integer.toString(startBit);
	    stringHi = Integer.toString(endBit);
	    lowLen = fm.stringWidth(stringLo);
	    highLen = fm.stringWidth(stringHi);
	    maxLabel = Math.max(lowLen, highLen);
	} 
	switch(labelStyle) {
	case LABEL_ALL:
	    chaX = 0;
	    chaY = 0;
	    chaW = rec.width;
	    chaH = fontH;
	case LABEL_VALUE:
	case LABEL_AXIS:
	    if (direction == DIR_UP || direction == DIR_DOWN) { 
		axisX = 3;  
		axisY = frameW + chaH;
		axisW = maxLabel + 4;
		axisH = rec.height - chaH - 2*frameW;
	    }
	    else {   //DIR_RIGHT || DIR_LEFT
		axisW = rec.width - 2*frameW;
		axisH = fontH;
		axisX = frameW;
		axisY = rec.height - axisH - 2;
	    }
	default:        //case LABEL_NONE:
	    if (direction == DIR_UP || direction == DIR_DOWN) {
		fillX = axisW + frameW; 
		fillY = frameW + chaH;
		fillW = rec.width - 2*frameW - axisW;
		fillH = rec.height - chaH - 2*frameW;
	    }
	    else {        // DIR_RIGHT || DIR_LEFT
		fillX = frameW; 
		fillY = chaH + frameW;
		fillW = rec.width - 2*frameW;
		fillH = rec.height - 2*frameW - chaH - axisH;
	    }
	}
    }


    protected void paintBackground (Graphics g) {
	initGraphics();
	g.setColor(getBackground());
	g.fill3DRect(0, 0, getSize().width, getSize().height, true);
	switch(labelStyle) {
	case LABEL_ALL:
	    drawName(g);
	case LABEL_VALUE:
	case LABEL_AXIS:
	    drawAxis(g);
	default:
	    g.setColor(getBackground());
	    g.fill3DRect(fillX - frameW, fillY - frameW,
			 fillW + 2*frameW, fillH + 2*frameW, true);
	}
    }


    protected void paintForeground (Graphics g) {
 	int bVal = value;
	//System.out.println("bVal="+bVal+", "+Integer.toBinaryString(bVal));
	bVal <<= 31 - endBit;

	//--- calculate the cellX & cellY ------------------------
	cellX = fillX;   cellY = fillY;
	cellW = fillW;   cellH = fillH;
	for (int i=0; i<cellCount; i++) {
	    switch (direction) {
	    case DIR_UP:
		cellY = fillY + (int)(i*((float)fillH/cellCount));
		cellH = (int)((i+1)*((float)fillH/cellCount)) - 
		        (int)(i*((float)fillH/cellCount));
		break;
	    case DIR_DOWN: 
		cellY = fillY+(int)((cellCount-1-i)*((float)fillH/cellCount));
		cellH = (int)((cellCount-i)*((float)fillH/cellCount)) -
		        (int)((cellCount-1-i)*((float)fillH/cellCount));
		break;
	    case DIR_RIGHT:
		cellX = fillX+(int)((cellCount-1-i)*((float)fillW/cellCount));
		cellW = (int)((cellCount-i)*((float)fillW/cellCount)) - 
                        (int)((cellCount-1-i)*((float)fillW/cellCount));
		break;
	    default:  //DIR_LEFT
		cellX = fillX+(int)(i*((float)fillW/cellCount));
		cellW = (int)((i+1)*((float)fillW/cellCount)) - 
                        (int)(i*((float)fillW/cellCount));
	    }
	    //--- draw cell -------------------
	    if (bVal < 0 ) {
		bitVal[endBit-i] = true;
		if(colorMode==CLRMOD_ALARM)  g.setColor(DEFAULT_ALARM_COLOR);
		else                         g.setColor(getForeground());	
		g.fill3DRect(cellX, cellY, cellW, cellH, true);
	    }
	    else {
		bitVal[endBit-i] = false;
		g.setColor(getForeground());	
		g.fill3DRect(cellX, cellY, cellW, cellH, false);
	    }
	    bVal <<= 1;
	}
    }

    protected void drawAxis(Graphics g) 
    {	
	int strLoX, strLoY, strHiX, strHiY = 0;
	g.setColor(fgColor);
	g.setFont(font);
	if (direction == DIR_UP || direction == DIR_DOWN) {
	    strHiX = axisX;
	    if(direction == DIR_UP)  strHiY = axisY + fontH;
	    else                     strHiY = axisY + axisH - 2;
	    strLoX = strHiX;
	    if(direction == DIR_UP)  strLoY = axisY + axisH - 2;
	    else                     strLoY = axisY + fontH;
	}
	else {
	    if (direction == DIR_RIGHT) strLoX = axisX;
	    else                        strLoX = axisX + axisW - lowLen;
	    strLoY = axisY + axisH;
	    if (direction == DIR_RIGHT) strHiX = axisX + axisW - highLen;
	    else                        strHiX = axisX;
	    strHiY = strLoY;
	}
	g.drawString(stringLo, strLoX, strLoY );
	g.drawString(stringHi, strHiX, strHiY );
        drawArrow(g);
    }


    private void drawArrow(Graphics g) {
	int[] arrowX = new int[3];
	int[] arrowY = new int[3]; 
	int x1, y1, x2, y2 = 0;
	switch (direction) {
	case DIR_UP:
	    arrowX[0] = (int) (axisX + 0.08*axisW);
	    arrowY[0] = (int) (axisY + 0.5*axisH);
	    arrowX[1] = (int) (axisX + 0.3*axisW);
	    arrowY[1] = (int) (axisY + 0.45*axisH);
	    arrowX[2] = (int) (axisX + 0.5*axisW);
	    arrowY[2] = (int) (axisY + 0.5*axisH);
	    x1 = (int) (axisX + 0.3*axisW);
	    y1 = (int) (axisY + 0.5*axisH);
	    x2 = x1;
	    y2 = (int) (axisY+ 0.65*axisH);
	    break;
	case DIR_DOWN:
	    arrowX[0] = (int) (axisX + 0.1*axisW);
	    arrowX[0] = (int) (axisX + 0.08*axisW);
	    arrowY[0] = (int) (axisY + 0.5*axisH);
	    arrowX[1] = (int) (axisX + 0.3*axisW);
	    arrowY[1] = (int) (axisY + 0.55*axisH);
	    arrowX[2] = (int) (axisX + 0.5*axisW);
	    arrowY[2] = (int) (axisY + 0.5*axisH);
	    x1 = (int) (axisX + 0.3*axisW);
	    y1 = (int) (axisY + 0.5*axisH);
	    x2 = x1;
	    y2 = (int) (axisY+ 0.35*axisH);
	    break;
	case  DIR_RIGHT:
	    arrowX[0] = (int) (axisX + 0.5*axisW);
	    arrowY[0] = (int) (axisY + 0.38*axisH);
	    arrowX[1] = (int) (axisX + 0.55*axisW);
	    arrowY[1] = (int) (axisY + 0.6*axisH);
	    arrowX[2] = (int) (axisX + 0.5*axisW);
	    arrowY[2] = (int) (axisY + 0.8*axisH);
	    x1 = (int) (axisX + 0.5*axisW);
	    y1 = (int) (axisY + 0.6*axisH);
	    x2 = (int) (axisX + 0.35*axisW);
	    y2 = y1;
	    break;
	default:
	    arrowX[0] = (int) (axisX + 0.5*axisW);
	    arrowY[0] = (int) (axisY + 0.38*axisH);
	    arrowX[1] = (int) (axisX + 0.45*axisW);
	    arrowY[1] = (int) (axisY + 0.6*axisH);
	    arrowX[2] = (int) (axisX + 0.5*axisW);
	    arrowY[2] = (int) (axisY + 0.8*axisH);
	    x1 = (int) (axisX + 0.5*axisW);
	    y1 = (int) (axisY + 0.6*axisH);
	    x2 = (int) (axisX + 0.65*axisW);
	    y2 = y1;
	    break;
	}
	g.drawPolyline(arrowX, arrowY, 3);
	g.drawLine(x1, y1, x2, y2);
    }


    private void drawName(Graphics g) 
    {
	String chaName = deviceName+"."+propertyName;
	String tempName = chaName;
	int chanLen = fm.stringWidth(tempName);
	int i = 2;
	// reduce the name string to fit the space
	while (chanLen > chaW) {
	    tempName = tempName.substring(0,tempName.length()-i);
	    i++;
	    chanLen = fm.stringWidth(tempName);
	}
	int stringW = fm.stringWidth(tempName);
	int stringX = chaX + (int)((chaW - stringW)/2);
	int stringY = chaY + chaH - 2;   
	g.setColor (fgColor);
	g.drawString (tempName, stringX, stringY);		
    }
    
    /**
     *  MouseListener interface
     */
    public void mouseEntered  (MouseEvent me) { }
    public void mouseExited  (MouseEvent me)  { }
    public void mousePressed (MouseEvent me)  { }
    public void mouseClicked (MouseEvent me) { }

    /**
     * Handles mouse release.
     * Mouse up on the bit cell, change the 1 to 0 or 0 to 1.
     */
    public void mouseReleased (MouseEvent me)
    {
       if (!controllable)  return;
       int x = me.getX();
       int y = me.getY();
       int val = value;
       int bitNumber = findBitNumber(x, y);
       if (bitNumber >= startBit && bitNumber <= endBit) {
	   //System.out.println("Bit["+bitNumber+"]="+bitVal[bitNumber]);
	   System.out.println("before val ="+val+" "+Integer.toBinaryString(val));
	   if(bitVal[bitNumber])  val -= 1<<bitNumber;
	   else                   val += 1<<bitNumber;
	   System.out.println("after val = "+val+" "+Integer.toBinaryString(val));
       }
       tdata.insert(VAL_TAG, val);
       try {
	   device.set (propertyName, tdata, (DeviceListener)null);
	   device.getControlSystem().flush();
	   System.out.println("send new value ("+val+")");
       } catch (Exception ex) {
	   System.out.println ("failed to send value\n"+ex);
       }
       value = val;
       repaint();
    }

    private int findBitNumber(int x, int y) {
	int cY, cH, cX, cW, i;
	boolean found = false;
	int returnNum = -1;
	if (x<fillX || x>fillX+fillW || y<fillY ||y>fillY+fillH)
	    return returnNum;	

	for (i=0; i<cellCount&&!found; i++) {
	    switch (direction) {
	    case DIR_UP:
		cY = fillY + (int)(i*((float)fillH/cellCount));
		cH = (int)((i+1)*((float)fillH/cellCount)) - 
		     (int)(i*((float)fillH/cellCount));
		if (y > cY && y < cY+cH)   found = true;
		break;
	    case DIR_DOWN: 
		cY = fillY+(int)((cellCount-1-i)*((float)fillH/cellCount));
		cH = (int)((cellCount-i)*((float)fillH/cellCount)) -
		     (int)((cellCount-1-i)*((float)fillH/cellCount));
		if (y > cY && y < cY+cH)   found = true;
		break;
	    case DIR_RIGHT:
		cX = fillX+(int)((cellCount-1-i)*((float)fillW/cellCount));
		cW = (int)((cellCount-i)*((float)fillW/cellCount)) - 
		     (int)((cellCount-1-i)*((float)fillW/cellCount));
		if (x > cX && x < cX + cW)   found = true;
		break;
	    default:  //DIR_LEFT
		cX = fillX+(int)(i*((float)fillW/cellCount));
		cW = (int)((i+1)*((float)fillW/cellCount)) - 
		     (int)(i*((float)fillW/cellCount));
		if (x > cX && x < cX + cW)   found = true;
	    }
	}
	if (found) returnNum = (endBit-i+1);
	return returnNum;
    }

 
    public void deviceChanged(DeviceEvent event) {
	if(firstCall) {
	    firstCall = false;
	    gotLimits = true;
	    startMonitor();
	}
	try { value = event.getData().getDataEntry("value").intValue();	} 
	catch (Exception e) { e.printStackTrace(); }  
	repaint();
    }
    
} // end CBits 










