/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  08/08/00:  Release version 0.4 (Initial version). 
 *
 */
 
package adm;

import java.io.*;   
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

import adm.event.*;
import cdev.*;
import cdev.data.*;

/**
 * CCartesian is a container widget that holds 1 to 2 traces. It does not
 * implement DeviceListener but receives a trace event from each trace when 
 * the new data point is available. The traces in the cartesian plot share
 * one common x-axis (first trace) but not the y-axis. The first trace has 
 * a the left side y-axis and the second trace has the right side y-axis. 
 * Each trace can have different x-axis range style, y-axis range style, 
 * trace style and trace color. 
 * <BR><BR>
 * If connected device - property pair contains array of data, the new data 
 * points will be plotted and the old points will be erased for each update. 
 * Otherwise the specified maximum number of update points (default of 30) 
 * will be plotted in the chart. When the next data point is received, the 
 * oldest point will be erased from the graph if there is more than maximum 
 * number of the data. This number can be reset using setMaxPoint method. 
 * The cartesian may have a title if setTitle is called. 
 * <BR><BR>
 * Each trace in the plot is defined in the CTrace class. It connects to 
 * two device/properties with a simultaneous data update. One is the value 
 * of the x-axis and another is the y-axis. Time stamps are used to trigger 
 * the trace event. This means that only a data pair with the same time 
 * stamp will be sent to CCartesian and plotted on the chart. CTrace defines 
 * many trace properties. You may specify one of three axes (x-axis or y-axis) 
 * range styles, RANGE_AUTO, RANGE_STATIC, RANGE_FROM_PROPRTY with 
 * setRangeStyle. The default range style for all axes are RANGE_FROM_PROPRTY.
 * Call setTraceColor method to reset the default trace color (green) and 
 * setTraceStyle to select STYLE_POINT or STYLE_LINE (default) as the trace 
 * style. CTrace class also provides setX(Y)DisplayLo(Hi) methods to specify
 * the display range. When any of these methods is called, the range style 
 * for that axis will be forced to RANGE_STATIC. If the range style is set 
 * to RANGE_STATIC and range low and/or range high are not set, the default 
 * display range (0 for low and 100 for high) will be used. 
 *
 *
 * @version   0.4  09/17/00
 * @author    Ying Chen, Chip Watson
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.4
 *
 */
public class CCartesian extends Panel implements CConstants {

    static final int RADIUS = 3;
    static final int INSETS = 8;
    private int maxPoint = 30;
    private String title = null;
    private int dataLength = 1;
    private CTrace[] traces = new CTrace[2];
    private CTraceListener tListener = null;
    private Vector[] data = {new Vector(maxPoint), new Vector(maxPoint)};
    private Vector[] points = {new Vector(maxPoint), new Vector(maxPoint)};

    //--- used for double-buffering ------------------------------
    protected Dimension offDimension;     // double-buffer size
    protected Image     offImage;         // off-screen buffer
    protected Graphics  offGraphics;      // off-screen paint handle

    //--- graphics data --------
    private int plotX, plotY, plotW, plotH;
    private double[] xLo = {0, 0}, xHi = {100, 100};
    private double[] yLo = {0, 0}, yHi = {100, 100};
    private boolean[] got_new_point = {false, false};
    private Axis X, Y0, Y1;
    private Font font;
    private FontMetrics fm;


    /**
     * Null Constructor
     */
    public CCartesian () { }


   /**
     * Constructor with parameters
     * @params trace    the trace appear on the plot (this trace
     *                  Y axis will on the left)
     */
     public CCartesian (CTrace trace) {
	 setTrace(0, trace);
     }

    /**
     * Constructor with parameters
     * @params trace1    the first trace appear on the plot
     *                   (its Y axis will on the left)
     * @params trace2    the second trace appear on the plot
     *                   (its Y axis will on the right)
     */
     public CCartesian (CTrace trace1, CTrace trace2) {
	 setTrace(0, trace1);
	 setTrace(1, trace2);
     }

    /**
     * Add a new trace to plot. The Y axis of the new trace is left
     * if which is 0 and right if which is 1.
     * @params which     must equal 0 or 1 
     * @params newTrace  the new trace class to appear on the plot
     */
    public void setTrace (int which, CTrace newTrace) {
	if (which < 0 || which > 1)  return;
	if (which == 1 && traces[0] == null) return;
	if (tListener == null) tListener = new CTraceListener() {
	    public void receivedNewData (CTraceEvent de) {
		dataLength = de.getDataLength();
		Double[] val;
		int j;  //set to 0/1 depend on which generate event 
		if((CTrace)de.getSource() == traces[0]) j = 0;
		else                                    j = 1;
		if (dataLength == 1) {
		    val = de.getDoubleValue(0);
		    //System.out.println("new X="+val[0].doubleValue()+", Y="+
		    //	       val[1].doubleValue()+" from traces"+j);
		    if (data[j].size() == maxPoint) 
		    data[j].removeElementAt(0);
		    data[j].addElement(val);
		}
		else {
		    data[j] = new Vector(dataLength);
		    for (int i = 0; i < dataLength; i++) {
			val = de.getDoubleValue(i);
			data[j].addElement(val);
		    }		    
		}
		got_new_point[j] = true;
		if (traces[j].redraw)  offGraphics = null;
		repaint();
	    }
	};

	traces[which] = newTrace;
	traces[which].addTraceListener(tListener);
	if (which == 0) 
	    traces[which].setYAxisPosition(AXIS_LEFT);
	else 
	    traces[which].setYAxisPosition(AXIS_RIGHT);
    }

    /**
     * Add trace to the cartesian plot.
     * @params trace   the new trace to appear on the plot.
     */
    public void addTrace (CTrace trace) {
	if (traces[0] == null)  setTrace(0, trace);
	else if (traces[1] == null)  setTrace(1, trace);
	else { System.out.println("More than two trace added"); }
    }


    /**
     * Overwrite container's add method for Viual tools
     * @params trace   the new trace to appear on the plot.
     */
    public Component add (CTrace trace) {
	addTrace(trace);
	return trace;
    }


    /**
     * Set up the title string appear on the top of the plot.
     * Default is null (noe title). 
     */
    public void setTitle (String aTitle) {
	title = aTitle;
	offGraphics = null;    
    }
	
    /**
     * Return the title string. There is no title if it is null.
     */
    public String getTitle () { return title; }

    /**
     * Set up max data point shown in the plot. When tha data
     * points is more than the max point, the oldest point will
     * be erased.
     */ 
    public void setMaxPoint (int num) { 
	maxPoint = num; 
	data[0] = new Vector(maxPoint);
	data[1] = new Vector(maxPoint);
	offGraphics = null;    
    }

    /**
     * Return max point shown in the plot
     */
    public int getMaxPoint () { return maxPoint; }

    /**
     * Over write the parant setBounds method to force draw the
     * background again when user resize the widget
     */
    public void setBounds(int x, int y, int w, int h) {
	super.setBounds(x, y, w, h);
        offGraphics = null;    
    }

    /**
     * Over write the parant setSize method to force draw the
     * background again when user resize the widget
     */
    public void setSize(int w, int h) {
	super.setSize(w, h);
	offGraphics = null; 
    }

    /**
     * Over write the Canvas.update to only call this.paint
     * (this is necessary since the Canvas.update will erase 
     * the background)
     */
    public final void update (Graphics g) {
	paint(g);
    }

    public void paint(Graphics g) {
	if (isShowing()) {
	    if (offGraphics == null) {
		setupOffscreen();
		paintBackground (offGraphics);
	    }
	    paintForeground (offGraphics);
	    g.drawImage(offImage, 0, 0, null);
	}
    }

   
    private void paintForeground (Graphics g) {
	if (data[0].size() == 0) return;
	if (dataLength == 1) {
	    if (got_new_point[0])  convertToPoint(0);
	    if (got_new_point[1])  convertToPoint(1);   
	}
	else {
	    if (got_new_point[0])  convertAllPoints(0);
	    if (got_new_point[1])  convertAllPoints(1); 
	} 
	//--- erase plot area and plot both trace --------
	g.setColor(getBackground());
	g.fillRect(plotX+1, plotY+1, plotW-2, plotH-2);
	drawTrace(0, g);
	if (traces[1] != null) drawTrace(1, g);
    }
    

    /**
     * Set up offscreen graphics.  This is needed for both 
     * a new screen paint and a resize paint, or anything which
     * changes the widget's background (color, font, etc.).
     */
    protected void setupOffscreen() {
        offDimension = getSize();
        offImage = createImage(offDimension.width, offDimension.height);
        offGraphics = offImage.getGraphics();
        offGraphics.setColor(getBackground());
        offGraphics.fillRect(0, 0, offDimension.width, offDimension.height);
    }

    /*
     * this is only need be called at first time
     */
    private void paintBackground (Graphics g) { 
	initGraphics();
	convertAllPoints(0);
	g.setColor(getBackground());
	g.fill3DRect(0, 0, getSize().width, getSize().height, true);
	if (title != null)   drawTitle(g);
	X.drawAxisTick (g);
	Y0.drawAxisTick (g);
	if (traces[1] != null) Y1.drawAxisTick (g);
    }


    private void initGraphics () {
	//--- set up the trace axis parameters ----------
	if (traces[0] == null) return;
	X = traces[0].getXAxis();
	Y0 = traces[0].getYAxis();
	traces[0].setXAxis(X);
	traces[0].setYAxis(Y0);
	xLo[0] = traces[0].getXDisplayLo();
	xHi[0] = traces[0].getXDisplayHi();
	if (!traces[0].gotRange()) offGraphics = null;
	System.out.println("CCartesian: xLo="+xLo[0]+", xHi="+xHi[0]);
	yLo[0] = traces[0].getYDisplayLo();
	yHi[0] = traces[0].getYDisplayHi();
	//System.out.println("CCartesian: y0Lo="+yLo[0]+", y0Hi="+yHi[0]);
	X.setDisplayRange(xLo[0], xHi[0]);
	Y0.setDisplayRange(yLo[0], yHi[0]);
	if (traces[1] != null) {
	    Y1 = traces[1].getYAxis();
	    traces[1].setYAxis(Y1);
	    traces[1].setXAxis(X);
	    yLo[1] = traces[1].getYDisplayLo();
	    yHi[1] = traces[1].getYDisplayHi();
	    if (!traces[1].gotRange()) offGraphics = null;
	    System.out.println("CCartesian: y1Lo="+yLo[1]+", y1Hi="+yHi[1]);
	    Y1.setDisplayRange(yLo[1], yHi[1]);
	}

	//--- set up graphics parameters ---------------
	Dimension rec = getSize ();
	font = getFont();  
	fm = Toolkit.getDefaultToolkit().getFontMetrics(font);
	plotX = Y0.getWidth() + INSETS;
	if (traces[1] == null) 
	    plotW = rec.width - plotX - 2*INSETS;
	else 
	    plotW = rec.width - plotX - Y1.getWidth() - 2*INSETS;
	if (title == null)  plotY = INSETS;
	else                plotY = 2*INSETS + fm.getHeight();
	plotH = rec.height - X.getHeight() - INSETS - plotY;
	//System.out.println("CCartesin::plot (x="+plotX+", y="+plotY+
	//		   ", w="+plotW+", h="+plotH+")");
	X.setAxisLength(plotW);
	X.setLocation(plotX, plotY+plotH);
	X.initGraphics ();
	Y0.setAxisLength(plotH);
	Y0.setLocation(INSETS, plotY);
	Y0.initGraphics ();
	if (traces[1] != null) {
	    Y1.setAxisLength(plotH);
	    Y1.setLocation(plotX+plotW, plotY);
	    Y1.initGraphics ();
	}
    }


    /* 
     * convert the new data to point and add to point vector
     */
    private void convertToPoint (int j) {
	Double[] val = (Double[])data[j].elementAt(data[j].size()-1);
	Point p = new Point();
	p.x = plotX+(int)(((val[0].doubleValue()-xLo[0])*plotW)/(xHi[0]-xLo[0]));
	p.y = plotH+plotY-(int)(((val[1].doubleValue()-yLo[j])*plotH)/(yHi[j]-yLo[j]));
	if (points[j].size() == maxPoint)  
	    points[j].removeElementAt(0);
	points[j].addElement(p);
	got_new_point[j] = false;
    }

    /*
     * For auto range. When the displayHi/Lo are changed, the points
     * in points vetctor must be recaculated using the data vector.
     */
    private void convertAllPoints (int j) {
	//Vector v = new Vector(maxPoint);
	Vector v = new Vector(data[j].size());
	for (int i=0; i<data[j].size(); i++) {
	    Double[] val = (Double[])data[j].elementAt(i);
	    Point p = new Point();
	    p.x = plotX+(int)(((val[0].doubleValue()-xLo[0])*plotW)/(xHi[0]-xLo[0]));
	    p.y = plotH+plotY-(int)(((val[1].doubleValue()-yLo[j])*plotH)/(yHi[j]-yLo[j]));
	    v.addElement(p);
	}
	points[j] = v;
	traces[j].redraw = false;
    }

    /*
     * draw trace points (draw evevy points for every updates)
     */
    private void drawTrace (int which, Graphics g) {
	g.setColor(traces[which].getTraceColor()); 
	g.setClip(plotX+1, plotY+1, plotW-2, plotH-2);
	for (int i=0; i<points[which].size(); i++) {
	    Point p = (Point) points[which].elementAt(i);
	    int traceStyle = traces[which].getTraceStyle();
	    switch (traceStyle) {
	    case STYLE_POINT:
		g.fillOval(p.x-RADIUS, p.y-RADIUS, 2*RADIUS, 2*RADIUS); 
		break;
	    case STYLE_FILLUNDER:
		System.out.println("not implement");
		break;
	    default:    // case STYLE_LINE:
		g.fillOval(p.x-2, p.y-2, 4, 4); 
		if (i > 0) {
		    Point pp = (Point) points[which].elementAt(i-1);
		    g.drawLine(pp.x, pp.y, p.x, p.y);
		}
	    }
	}
    }
    
    /*
     * draw the title string on the top of plot
     */
    private void drawTitle (Graphics g) {
	int stringW = fm.stringWidth(title);
        int stringX = (int)((getSize().width - stringW)/2);
	int stringY = INSETS + fm.getHeight();
	g.setColor(Color.black);
	g.drawString(title, stringX, stringY);
    }

}
