/**
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  10/06/00:  Created.
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

import cdev.*;
import cdev.data.*;


/**
 * CCheckbox is an awt-like Checkbox widget. It only works with the
 * device/property with boolean value or the two state Enum type. 
 * If the data type is not right, the checkbox label will be set to
 * "Error!!!" and there is no action occur.
 * <BR><BR>
 * By default the widget has no label. Call the setLabel(String label)
 * method, can add the label to the CCheckbox widget. 
 * <BR><BR>
 * The device name and property name must be set using setDeviceName
 * and setPropertyName or the correct constructor 
 * CCheckbox(String dname, String pname).
 * <BR><BR>
 * If the current state is true or 1 (second Enum state), the checkbox 
 * is be checked (on). Otherwise is not be checked (off). Clicking on 
 * check box widget set its state from "on" to "off," or from "off" to
 * "on." If the set operation failed, the checkbox will back to the 
 * original state. 
 *
 *
 * @version   1.0  10/06/00
 * @author    Ying Chen, Chip Watson,
 * @since     adm1.0
 */
public class CCheckbox extends Checkbox 
    implements CInterface,  DeviceListener, ItemListener {

    static final int VAL_TAG = Data.convertKey("value");
    static final String BLANK = "";
    private static DeviceContext defContext = new DeviceContext();
    private static DeviceContext limitsContext;
    static {
	Data limitsContextData = new Data();
	limitsContextData.insert("displayHigh", 1);
	limitsContext = new DeviceContext(limitsContextData);
    }

    private Data tdata = new Data();
    private boolean active = false;
    private Device device = null;
    private String deviceName = null;
    private String propertyName = null;
    private String lastDeviceName = null;
    private String lastPropertyName = null;
    private String label = null;

    private boolean firstCall = true; 
    private boolean dataTypeError = false;
    private String[] tags;
    private String value;
    private DeviceListener setListener = null;

    /**
     * Constructor
     */
    public CCheckbox()
    {
	super();
	addItemListener(this);
    }


    /**
     * Constructor  with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CCheckbox (String dName, String pName)  {
	this();
	setDeviceName(dName);
	setPropertyName(pName);
	super.setLabel(pName);
    }

    public void setLabel (String newLabel) {
	label = newLabel;
	super.setLabel(newLabel);
    }

    public String getLabel () { return label; }


   /**
     * Set limitsContext to get all meta-data at first time
     */
    synchronized protected void getFirstValue () 
    {
       try {
	  if (deviceName == null || propertyName == null)   return;
          if (!deviceName.equals(lastDeviceName)) {
             device = new Device(deviceName);
             lastDeviceName = deviceName;
             }
          device.setContext (limitsContext);
	  device.get(propertyName, this);
          firstCall = true; 
          }
       catch(Exception e) { e.printStackTrace(); }
     }
     

    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setDeviceName(String newName) 
    {
       newName = CUtility.makeValidName(newName);
       if (!CUtility.testNameMatch(deviceName, newName)) {
          stopMonitor();
          deviceName = newName;
          getFirstValue ();
       } 
    }
    

    /**
     * Return the device name to which this component is connected.
     * If device name is not set, return empty string "". 
     */
    public String getDeviceName()
    {
       return (deviceName==null)?BLANK:deviceName;
    }


    /**
     * Set device property name 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setPropertyName(String newName)
    {
	newName = CUtility.makeValidName(newName);
        if (!CUtility.testNameMatch(propertyName, newName)) {
           stopMonitor();
           propertyName = newName; 
           getFirstValue ();        
        }
    }


    /**
     * Return the device property name to which this component is connected.
     * If property name is not set, return empty string "". 
     */
    public String getPropertyName()
    {
	return (propertyName==null)?BLANK:propertyName;
    }
	

    private void stopMonitor() 
    {
        if (active) {
          try {
	     active = false;
	     device.monitorOff(propertyName,this);
	  } 
	  catch (IOException e) {
	     e.printStackTrace();
	  }
       }
    }
	
    private void startMonitor() 
    {
	try {
	    if (active) stopMonitor();
	    if (deviceName == null || propertyName == null) return;
	    if (!deviceName.equals(lastDeviceName)) {
		device = new Device(deviceName);
		lastDeviceName = deviceName;
	    }    
	    device.setContext (defContext);
	    device.monitorOn(propertyName,this);
	    lastPropertyName = propertyName;
	    active = true;
	} 
	catch (IOException e) {
	    active = false;
	    e.printStackTrace();
	}
    }

    /**
     * Set the change back to control system and monitor the result.
     * if fail it will back to original state
     */
    public void itemStateChanged(ItemEvent evt) {
	if (setListener==null) setListener = new DeviceListener() {
	    public void deviceChanged(DeviceEvent evt) {
		if (evt.getError() == null) {
		    System.out.println("success to set value");
		} else {
		    System.out.println("failed to set the value");
		    setState(getTagState(value));
		}
	    }
	};

	if (dataTypeError) {
	    setState(!getState());
	}
	else {
	    String val;
	    if (getState()) val = tags[1];
	    else            val = tags[0];
	    tdata.insert(VAL_TAG, val);
	    //System.out.println(tdata);
	    try {
		device.set (propertyName, tdata, setListener);
		device.getControlSystem().flush();
	    } catch (Exception ex) {
		System.out.println ("fail to set value ("+val+")");
		setState(getTagState(value));
	    }
	}
    }
       
    private void printErrorMessage () {
	dataTypeError = true;
	setLabel("Error!!!");
	System.out.print("Invalid "+deviceName+"."+propertyName);
	System.out.println(" (data type is not boolean 0/1)");
    }

    /**
     * First call back to check the device-property type, there is no 
     * action occurs if none boolean data type.
     * For the monitorOn call back, reset the checkbox state and reset
     * the label if user don't set the label
     */
    public void deviceChanged (DeviceEvent event) {
	if (firstCall) {    // from get call
	    firstCall = false;
	    try {
		DataEntry de = event.getData().getDataEntry("value");	
		if (de.getValue() instanceof String[]) {
		    tags = de.stringArray();
		    if (tags.length != 2) printErrorMessage();
		    else                  startMonitor();
		}
		else { 
		    printErrorMessage();
		} 		
	    } catch (Exception e) { e.printStackTrace(); }   
	}
	else {    // from monitorOn
	    if (!dataTypeError) {
		value = event.getData().getDataEntry("value").stringValue();
		if (label != null) setLabel(label);
		else               super.setLabel("");
		if (isOneOfTag(value)) 
		    setState(getTagState(value));
	    }
	}
    }


    private boolean isOneOfTag (String tagName) {
	return (tagName.equals(tags[0]) || tagName.equals(tags[1]));
    }

    private boolean getTagState(String tagName) {
	if (tagName.equals(tags[0])) return false;
	else                    return true;
    }


}  // end CCheckbox





