/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * 
 * Revision History:
 * 06/24/00: Release 0.3 (initial version)
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;


/**
 * CJavaMenuItem is a subclasses of awt.MenuItem. It is a invisible beans 
 * and must be added to a CMenuButton widget. Unlike awt.MenuItem, the 
 * action event is already built into this widget. The user does not need 
 * to implement the ActionListener. 
 * <BR><BR>
 * CJavaMenuItem is used to launch a java application within the same 
 * virtual machine. The application is required to provide a static method 
 * named execute (String params[]). The name of this application must be 
 * set through the setClassName (String javaName) method. If the params[] 
 * is used in the execute method, it must be set using setParams 
 * (String[] params) method. 
 * <BR><BR>
 * The default label for CJavaMenuItem is String "[J]". If user set the label, 
 * the "[J] "+label will be the real label to distinguish them from 
 * awt.MenuItem. 
 *
 *
 * @version  0.3  05/17/00
 * @author   Ying Chen, Chip Watson
 *            
 * @since    adm0.3
 *
 */
public class CJavaMenuItem extends MenuItem implements ActionListener {
    
    static final String METHOD = "main";  
    private String className = null;
    private String[] params = null;
    private String label = "";
 
    /**
     * Null constructor. 
     */
    public CJavaMenuItem () {
	this("");
    }

    /**
     * Constructor  with label parameter
     *
     * @param label    label string 
     */
    public CJavaMenuItem (String label) {
	setLabel(label);
	addActionListener(this);
    }

    /**
     * Set the label string. The real label will be "[J] "+label.
     *
     * @param label      the label String 
     */
    public void setLabel (String label) {
	this.label = label;
	super.setLabel("[J] "+label);
    }
    
    /**
     * Return the label string (not include "[J] ").
     */
    public String getLabel() { return label; }
     
    /**
     * Set up the java class name launched by this CJavaMenuItem.
     * The user must call this method.
     *
     * @param cName     the java class name need launched 
     */  
    public void setClassName (String cName) { className = cName; }

    /**
     * Return the java class name.
     */
    public String getClassName () { return className; }

    /**
     * Set up the parameters exexute method needed. It must called
     * if java class provide with parameters execute method.
     *
     * @param params[]   parameters of execute method (array of String)
     */
    public void setParams (String[] params) { this.params = params; }
 
    /**
     * Return parameters String array.
     */
    public String[] getParams () { return params; }

    /**
     * Launch the class by invoke the execute method in the user's
     * application.
     */
    public void actionPerformed (ActionEvent evt) {
	Class c = null;
	Method m = null;
	Object o = null;
	
	setEnabled(false);
	if (className == null) { 
	    System.out.println("no java class is set");
	    return; 
	}
	try {  c = Class.forName(className); }
	catch(ClassNotFoundException nce) {
	    System.out.println("no class ("+className+") is found");
	    return;
	}   

	Class[] paramTypes = new Class[1];
	paramTypes[0] = (new String[0]).getClass();
	try { m = c.getMethod(METHOD, paramTypes); }
	catch (NoSuchMethodException nme) {
	    System.out.println("no method ("+METHOD+") is found");
	    return;
	}
	Object[] refParams = new Object[1];
	refParams[0] = params;
	
	try { 
	    o = m.invoke(null, refParams); 
	    setEnabled(true);
	}
	catch (Exception e) { 
	    System.out.println("failed to invoke ("+METHOD+") "+e);
	}
    }

} // end of CJavaMenuItem class
    
