/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 * Revision History:
 *   05/17/00:  Release version 0.3 (inital version)
 *
 */   

package adm;

import java.awt.*;
import java.awt.event.*;

/**
 * This class creates a stand-alone menu button. It does not need to be 
 * added to a menuBar. CMenuBotton looks like a button but works like a 
 * menu. When the button is pressed, a pull down menu will appear. 
 * <BR><BR>
 * Like awt.Menu, any menuItem, separator and subclass of menuItem such 
 * as JavaMenuItem and ShellMenuItem can be added to this widget. 
 * <BR><BR>
 * To distinguish from a normal awt.Button, a ">>" String is added on back
 * of the label set by the user. If the label is not set, the default label 
 * ">>" will appear on the CMenuBotton. 
 *
 *
 * @version   0.3  06/24/00
 * @author    Ying Chen, Chip Watson
 *            
 * @since     adm0.3
 *
 */
public class CMenuButton extends Button implements ActionListener {

    static final String SPECIAL_LABEL = " >> ";
    private String label;
    private PopupMenu popup = new PopupMenu();

    /**
     * Null constructor. 
     */
    public CMenuButton () {
	this("");
    }

    /**
     * Constructor  with label parameter
     *
     * @param label    label string 
     */
    public CMenuButton (String label) {
	super(label+SPECIAL_LABEL);
	this.label = label;
	addActionListener(this);
	super.add(popup);
    }

    /**
     * Return the label string (not include " >>").
     */
    public String getLabel() { return label; }

    /**
     * Set the label string. The real label will be label+" >>".
     *
     * @param label      the label String 
     */
    public void setLabel(String label) {
	super.setLabel(label+SPECIAL_LABEL);
	this.label = label;
    }

    /**
     * Adds the specified menu item to this menu. If the menu item has 
     * been part of another menu, remove it from that menu. 
     */
    public synchronized MenuItem add (MenuItem mi) { 
	if (mi == null) return null;
	return popup.add(mi); 
    }

    /**
     * Adds a separator line, or a hypen, to the menu at the current 
     * position. 
     */
    public void addSeparator() { popup.addSeparator(); }

    /**
     * Inserts a menu item into this menu at the specified position. 
     */
    public synchronized void insert (MenuItem mi, int index) {
	if (mi == null || index < 0) return;
        popup.insert(mi, index);
    }

    /**
     * Removes the menu item at the specified index from this 
     * menuButton. 
     */
    public synchronized void remove (int index) {
	if (index < 0) return;
	popup.remove(index);
    }

    /**
     * Removes the specified menu item from this menuButton. 
     */
    public void remove (MenuComponent item) {
	popup.remove(item);
    }

    /**
     * Inserts a separator at the specified position.
     */ 
    public void insertSeparator(int index) {
	if (index < 0) return;
	popup.insertSeparator(index);
    }

    /**
     * Popup a popup menu at position (0, height of button)
     */
    public void actionPerformed (ActionEvent evt) {
	popup.show(this, 0, getSize().height);
    }
	
}

    
