/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  03/10/00:  Release version 0.1.
 *  04/17/00:  Release version 0.2.
 *  06/13/00:  Release version 0.3. (add the code to handle the Enum type)
 *
 */

package adm;

import java.awt.*;
import java.io.*;
import java.util.*;

import cdev.*;
import cdev.data.*;

/**
 * CRect awt component bean. It is used to monitor the state of a 
 * specific device property in a control system. The size, bound, 
 * foreground and background can be set like the awt component. 
 * <BR><BR>
 * CRect has the color mode property. When it is set to CLRMOD_ALARM, 
 * the color of the widget will change from foreground to yellow or red 
 * if the property is in warning or alarm state (determined by warning 
 * and alarm limits). When the foreground and the background color are 
 * the same as the container's, this widget will only appear when the 
 * device property is in the warning and alarm states. 
 * <BR><BR>
 * Since version 0.3 this widget can used to monitor the Enum type 
 * device - property. A color map (table) is used to represent the state 
 * when the color mode is CLRMOD_ALARM. Default colorMap contains 12 
 * colors, and uses color black for state above 12. User can overwrite 
 * this colorMap by calling the setColorMap (Color[] newColor). If the 
 * length of the newColor is less than the number of the Enum states, 
 * the color black will be used for the remaining states. 
 *
 *
 * @version   0.2  04/17/00
 * @author    Ying Chen, Chip Watson
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.1
 */
public class CRect extends CCanvas {

    static final Color C0 = Color.red;
    static final Color C1 = Color.orange;
    static final Color C2 = Color.yellow;
    static final Color C3 = new Color(176, 255, 176);
    static final Color C4 = Color.green; 
    static final Color C5 = new Color(0, 128, 64);
    static final Color C6 = Color.blue;
    static final Color C7 = new Color(0, 0, 160);
    static final Color C8 = new Color(128, 0, 255);
    static final Color C9 = new Color(64, 128, 128);
    static final Color C10 = new Color(128, 0, 0);      // brown
    static final Color C11 = new Color(128, 128, 128);  // gray
    static final int MINSIZE = 2;

    protected boolean enumType = false;
    protected boolean gotLimits = false;
    protected boolean firstCall = true;
    protected boolean firstPaint = true;

    protected String stateName = "Good";
    protected Hashtable colorTable = new Hashtable();
    protected Color[] colorMap={C0,C1,C2,C3,C4,C5,C6,C7,C8,C9,C10,C11}; 

    //--- define default value --------------------------
    protected double currVal;
    protected Color currColor = null;
    protected String label = null;
    protected int labelX, labelY;

    /**
     *  Null Constructor.
     */
    public CRect ()
    {
	this (MINSIZE, MINSIZE);
    }

    /**
     * Constructor with device name and property name
     *
     * @param dName    device name
     * @param pName    property name
     */
    public CRect(String dName, String pName)
    {
	this (MINSIZE, MINSIZE); 
	setDeviceName(dName);
	setPropertyName(pName);
    }

    /**
     * Constuctor.
     * @param w   the width of the component
     * @param h   the height of the component
     */
    public CRect (int w, int h)
    {
        this (w, h, Color.green, Color.gray);
    }


    /**
     * Constuctor.
     * @param w   the width of the component
     * @param h   the height of the component
     * @param fg  the foreground color of the component
     * @param bg  the background color of the component
     */ 
    public CRect (int w, int h, Color fg, Color bg) 
    {
	setSize (w, h);
	if (fg != null)  setForeground(fg);
	if (bg != null)  setBackground(bg);
    }
	
    /**
     * Set the label string appeared on the widget 
     * If null, there is no label on the widget (default)
     * @params newLabel  the label string
     */
    final public void setLabel (String newLabel) {
	label = newLabel;
    }

    /**
     * Return the label string. There is no lable if null returned
     */
    final public String getLabel () { return label; }

    /**
     * Set the color mode property
     * @param newColorMode   the new color mode, which may be one of:
     *             <OL>
     *             <LI>  CConstants.CLRMOD_STATIC
     *             <LI>  CConstants.CLRMOD_ALARM
     *             </OL>
     */
    final public void setColorMode (int newColorMode) {
	if (colorMode != newColorMode) 
	    colorMode = newColorMode;
    }     
    
    /**
     * Return the color mode property
     */
    final public int getColorMode () { return colorMode; }
 
    
    /**
     * Set the color map for the Enum type device property to overwrite
     * the default color array. The size of the color array must equal 
     * to the size of Enum string. The widget will use the black to 
     * replace the missing color. 
     *
     * @param colors   a array of colors              
     */
    final public void setColorMap (Color[] colors) { colorMap = colors; }

    /**
     * Get the color map for the Enum type device property
     */
    final public Color[] getColorMap () {
	/*
	Color[] tmpColors = null;
	if (colorTable.size() > 0) { 
	    for (int i=0; i<colorTable.size(); i++) {
		if (i<colorMap.length)  tmpColors[i] = colorMap[i];
		else                    tmpColors[i] = Color.black;
	    }
	}
	return tmpColors;
	*/
	return colorMap;
    }
	
    /**
     * draw Rectange and label if label is set
     * @param g   offGraphics object
     */
    public void paint (Graphics g) 
    {
	if (firstCall)  return; // don't draw before first call
	Dimension rec = getSize();    
	if (isShowing() && gotLimits) {
	    g.setColor(currColor);
	    g.fillRect(0, 0, rec.width, rec.height);
	    if (label != null)   drawLabel(g);
	}
    }


    private void drawLabel (Graphics g) 
    {
	if (getFont() == null) setFont(new Font("Dialog", Font.PLAIN, 12));
	FontMetrics fm = Toolkit.getDefaultToolkit().getFontMetrics(getFont());
	Dimension rec = getSize();    	
	int stringH = fm.getHeight();
	int stringW = fm.stringWidth(label);
	labelX = (int)((rec.width - stringW)/2);
	labelY = (int)((rec.height + stringH)/2);
	g.setColor(Color.black);
	g.drawString(label, labelX, labelY);
    }
    

    /**
     * To get the new value, to calculate the new color
     * and call repaint() to update the color.
     */
    public void deviceChanged (DeviceEvent evt) {
        if (firstCall)   {  // from first get call
	    findInfo (evt); 
	} else {            // from monitorOn	
	    try {
		if (enumType) {
		    stateName = evt.getData().getDataEntry("value").stringValue();
		    currColor = (Color) colorTable.get(stateName); 
		}
		else {
		    if (colorMode == CLRMOD_ALARM) {
			currVal = evt.getData().getDataEntry("value").doubleValue();
			//--- find out the currColor used in paint ---------
			if ((currVal >=alarmHi) || (currVal <=alarmLo)) 
			    currColor =  alarmColor;
			else if ((currVal >=warningHi) || (currVal <=warningLo)) 
			    currColor =  warningColor;
			else 
			    currColor =  getForeground();
		    }
		    else 
			currColor = getForeground();
		}
	    } catch (Exception e) { e.printStackTrace(); }
	}
	repaint();
    }
	
    /**
     * For the first call using the fat context to get the device
     * property infomation. Only called at first deviceChange method.
     */
    private void findInfo (DeviceEvent evt) {
	try {
	    DataEntry de = evt.getData().getDataEntry("value");
	    //--- for Enum type --------------------------
	    if (de.getValue() instanceof String[]) {
		enumType = true;
		String[] tags = de.stringArray(); 	    
		for (int i=0; i<tags.length; i++) {
		    if (i<colorMap.length) 
			colorTable.put(tags[i], colorMap[i]); 
		    else                    
			colorTable.put(tags[i], Color.black);
		}
		gotLimits = true;
	    }
	    else { //--- for non Enum type -----------------------
		enumType = false;
		try {
		    currVal=evt.getData().getDataEntry("value").doubleValue();
		    displayHi=evt.getData().getDataEntry("displayHigh").doubleValue();
		    displayLo=evt.getData().getDataEntry("displayLow").doubleValue();
		    gotLimits = true;
		} catch(Exception ee) { 
		    System.out.println("Failed to get display range");
		}
		
		//--- get the alarm and warning range ----------------------
		try {
		    alarmHi=evt.getData().getDataEntry("alarmHigh").doubleValue();
		    alarmLo=evt.getData().getDataEntry("alarmLow").doubleValue();
		}
		catch  (Exception e) {
		    alarmHi = displayHi;
		    alarmLo = displayLo;
		    System.out.println("Failed to get alarm range");
		}
		try {
		    warningHi=evt.getData().getDataEntry("warningHigh").doubleValue();
		    warningLo=evt.getData().getDataEntry("warningLow").doubleValue();
		}
		catch  (Exception e) {
		    warningHi = displayHi;
		    warningLo = displayLo;
		    System.out.println("Failed to get warining range");
		}
	    }
	    firstCall = false;		
	    startMonitor();
	}
	catch (Exception e) { e.printStackTrace(); }  
    }
    
} //end of CRect





