/* Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *  Revision History:
 *  04/17/00:  Release version 0.2. 
 *  09/06/00:  Release 0.4 (add isCommonAxis() method)
 *  11/07/00:  Release 1.0 (remove deprecated getCommonAxis())
 */

package	adm;

import	java.io.*;   
import	java.awt.*;
import	java.awt.event.*;
import	java.awt.image.*;
import  java.util.*;

import  cdev.*;
import  cdev.data.*;

/**
 * CStripChart is a container widget that holds up to 8 Pen instances. It 
 * does not implement DeviceListener but gets the data from each Pen at 
 * specified time intervals. The new data will be plotted on the right 
 * side of the strip chart and the old data will be shifted to the left. 
 * <BR><BR>
 * The device name and property name must be set for each Pen added to 
 * the CStripChart. The color of the Pen is optional. The default Pen color 
 * will be used if it has not been set. The default Pen color defined in 
 * CStripChart class is brown, orange, green, blue, purple, pink, cyan, 
 * navy for pen[0], pen[1], ..., pen[7]. 
 * <BR><BR>
 * The Pen in the CStripChart can share a common display range (y-axis) 
 * using method setCommonAxis (true). By default, all the pens have their 
 * own display range. 
 * <BR><BR>
 * Initially the information of the first Pen will be displayed on the y 
 * axis label and device/property name label in the color of the first Pen. 
 * The value and color of the label will change to a specific Pen if the 
 * user clicks on the plot line of that Pen. Due to a java 1.1.X bug, this 
 * feature only works with the jre 1.2. 
 * <BR><BR>
 * The methods setDisplayRange (int, int), setDisplayHi (int) and 
 * setDisplayLo (int) in the CStripChart class will force setting the range 
 * style to RANGE_STATIC and common axis to true. When using a common axis, 
 * the minimum display low and the maximum display high of all pens are 
 * used as the display range if the user does not specify the display 
 * range. The display range of each pen can be set individually using 
 * the setDisplayHi(double)/setDispalyLo(double) or 
 * setDisplayRange(double, double) methods in the Pen class. 
 * <BR><BR>
 * The y-axis is forced to a common axis when the range style is set to 
 * RANGE_AUTO.
 * 
 *
 * @version   0.2  04/15/00
 * @author    Ying Chen, Chip Watson
 *            (Based on work by Cuong Quach and Mr. John Winans)
 * @since     adm0.2
 */
public class CStripChart extends Panel 
    implements CConstants, MouseListener, Runnable {

    //--- define static constants --------------------------
    static final int MAX_PENS = 8;
    static final int MINSIZE = 20;
    static final int TICK_LEN = 5;  
    static final int GAP = 4;  
    static final int MIN_UPDATE_PERIOD = 2000;   // 2 sec.
    static final int DEFAULT_DX = 1;             // 2 pixles
    static final double BAD_VALUE = -11111.0;
    static final Color orange = Color.orange;
    static final Color green = Color.green;
    static final Color blue = Color.blue;
    static final Color purple = new Color(128, 0, 128);
    static final Color pink = Color.pink;
    static final Color cyan = Color.cyan;
    static final Color brown = new Color(128, 64, 0);
    static final Color navy =  new Color(0, 64, 128);

    //--- used for double-buffering ---------------------------
    private Dimension offDimension;   // double-buffer size
    private Image     offImage;       // off-screen buffer
    private Graphics  offGraphics;    // off-screen paint handle

    //--- define default value ------------------------------
    private Color   fgColor = Color.black;
    private Color   bgColor = Color.lightGray;
    private Font    font = new Font("Dialog", Font.PLAIN, 12);    
    private int     rangeStyle = RANGE_FROM_PROPERTY;
    private int     labelStyle = LABEL_NONE;
    private String  fmt = "%.3g";       
    private boolean commonAxis = false;
    //private boolean gotRange = false;
    private boolean redrawBg = false;

    //--- used for graphic plot -----------------------------
    private int faceX=0, faceY=0, faceW=0, faceH=0;
    private int graphX, graphY, graphW, graphH = 0;
    private int xAxisX, xAxisY, xAxisW, xAxisH = 0;
    private int yAxisX, yAxisY, yAxisW, yAxisH = 0;
    private int chaX, chaY, chaW, chaH = 0;
    private String stringYHi, stringYLo, stringXLabel;
    private int fontH, gapX, gapY = 0;
    //private FontMetrics fm = getFontMetrics(font); // for java 1.2
    private FontMetrics fm=Toolkit.getDefaultToolkit().getFontMetrics(font);

    private int strLoX, strLoY, strHiX, strHiY;   
    private double displayHi=100, displayLo=0;

    //--- used for pens ----------------------------------------
    private CPen[] pen = new CPen[MAX_PENS];
    private Vector[] history = new Vector[MAX_PENS];
    private Color[] defColor={brown,orange,green,blue,purple,pink,cyan,navy};
    private Color[] penColor = new Color[MAX_PENS];
    private double[] hi = new double[MAX_PENS];
    private double[] lo = new double[MAX_PENS];
    private double[] newValue = new double[MAX_PENS];
    private int[] oldY = new int[MAX_PENS];
    private int[] newY = new int[MAX_PENS];
    private int oldX, newX;
    private int currPen = 0, penNum = 0;
    private double[] prettyX = new double[3]; 
    private double[][] prettyY = new double[MAX_PENS][3]; 

    private Thread timer = null;
    private int updatePeriod = MIN_UPDATE_PERIOD;
    private int displayPeriod = 600;         // 600 sec or 10 min.
    private int dx = DEFAULT_DX;   
    private int totalXPoint;
    private double value, percent;

    /**
     * Null Constructor
     */
    public CStripChart () {
	setLayout(null);
	addMouseListener(this);
	start();
    }

    /**
     * Returns true if there are Pen on the stripChart available 
     * (max available Pen is 8). Otherwise returns false.
     */
    public boolean pensAvailable () {
	return (penNum < MAX_PENS)? true:false;
    }

    /**
     * Add a new pen to StripChart if there is a Pen available.
     * The user must set the device and property name of the Pen
     * before add to stripChart. The color of the Pen is option.
     * The stripChart will use a default Pen color if the color
     * of Pen is not be set in Pen class. 
     * @param newPen    a new CPen class
     */
    public void addPen (CPen newPen) {
	if (pensAvailable()) {
	    pen[penNum] = newPen;
	    penNum++;
	}
    }

    /**
     * Add a new Pen to StripChart if there is a Pen available.    
     * Specify the deviceName and propertyName it want to connect.
     * Use the default pen color and get the displayHi/displayLo
     * from the control system if the commonAxis is set to false. 
     *
     * @param dName    the device name for the pen to connect
     * @param pName    the property name for the pen to connect
     */
    public void addPen (String dName, String pName) {
	if (pensAvailable()) {
	    CPen newPen = new CPen(dName, pName);
	    pen[penNum] = newPen;
	    penNum++;
	}
    }

   /**    
    * Add a new Pen to StripChart if there is a Pen available.
    * Get the display range from control system if the commonAxis
    * is set to false.
    *
    * @param dName     the device name for the pen to connect
    * @param pName     the property name for the pen to connect
    * @param penColor  the color of the pen
    */
    public void addPen (String dName, String pName, Color penColor) { 
	if (pensAvailable()) {
	    CPen newPen = new CPen(dName, pName);
	    newPen.setPenColor(penColor);
	    pen[penNum] = newPen;
	    penNum++;
	}
    }

    /**    
     * Add a new Pen to StripChart if there is a Pen available.
     * Specify the deviceName/propertyName, Pen color and
     * display range value.
     *
     * @param dName     the device name for the pen to connect
     * @param pName     the property name for the pen to connect
     * @param penColor  the color of the pen
     * @param min       the display range Low
     * @param max       the display range High
     */
    public void addPen (String dName, String pName, Color penColor, 
			double min, double max) {
	if (pensAvailable()) {
	    CPen newPen = new CPen(dName, pName);
	    newPen.setPenColor(penColor);
	    newPen.setDisplayRange(min, max);
	    pen[penNum] = newPen;
	    penNum++;
	}
    }

    /**
     * Overwrite the Container's add for visual tool
     * @param newPens    a CPen object.
     */
    public Component add (CPen pen) {
        addPen(pen);
        return pen;
    }

    /**
     * Set the label style property of stripChart
     * @param newLab   the new label style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.LABEL_NONE
     *             <LI>  CConstants.LABEL_AXIS
     *             <LI>  CConstants.LABEL_ALL
     *             </OL>
     */
    public void setLabelStyle (int label) {
	labelStyle = label;
	offGraphics = null;
    }

    /**
     * Return the label style property
     */
    public int getLabelStyle () { return labelStyle; }


    /**
     * Set the range style property of stripChart
     * @param newRangeStyle  the new range style, which may be one of:
     *             <OL>
     *             <LI>  CConstants.RANGE_AUTO
     *             <LI>  CConstants.RANGE_FROM_PROPERTY
     *             <LI>  CConstants.RANGE_STATIC
     *             </OL>
     */
    public void setRangeStyle (int newRangeStyle) {
	rangeStyle = newRangeStyle;
	offGraphics = null;
    }

    /**
     * Return the range style property
     */
    public int getRangeStyle () { return rangeStyle; }

    /**
     * Set up commonAxis property. If commonAxis = true, all the Pen
     * shear one common axis (display range). Use the displayHi/displayLo
     * of Pen[0] if the common display range (displayHi/displayLo) are 
     * not set by user.
     */
    public void setCommonAxis (boolean b) {
	commonAxis = b;
	offGraphics = null;
    }
    
    /**
     * Return the commonAxis property. Shear same y axis if true.
     * Otherwise each Pen has its own display range.
     */
    public boolean isCommonAxis () { return commonAxis; }


    /**
     * Set the total display time period. This is the time range
     * displayed in the stripChart X axis. The unit of displayPeriod 
     * is seconds. The label of time axis (x axis) will shows minutes
     * if the displayPeriod >= 120 seconds.
     * In stripChart all Pens have same time axis (x axis).
     *
     * @param seconds    the total display period in seconds
     */
    public void setDisplayPeriod (int seconds) {
	displayPeriod = seconds;
	offGraphics = null;
    }

    /**
     * Return the displayPeriod in seconds.
     */
    public int getDisplayPeriod () {return displayPeriod; }

   /**
     * Set the display limits. The range style property will be 
     * forced to RANGE_STATIC. And the commonAxis will be forced 
     * to ture. The display range set in the Pen will be ignored.
     * 
     * @param lo   the display low
     * @param hi   the display high
     *
     * hi must be larger than lo otherwise nothing is changed.
     */
    public void setDisplayRange (double lo, double hi)
    {
	displayHi = Math.max(lo, hi);
	displayLo = Math.min(lo, hi);
	commonAxis = true;
	rangeStyle = RANGE_STATIC;
	offGraphics = null;
    }

    /**
     * Set the display high. The range style property
     * will be forced to RANGE_STATIC.
     * 
     * @param hi   the display high
     *
     * hi must be larger than displayLo otherwise nothing is changed.
     */
    public void setDisplayHi (double hi)
    {
	displayHi = hi;
	commonAxis = true;
	rangeStyle = RANGE_STATIC;
	offGraphics = null;
    }

    /**
     * Set the display low. The range style property
     * will be forced to RANGE_STATIC.
     * 
     * @param lo   the display low
     *
     * lo must be smaller than displayHi otherwise nothing is changed.
     */
    public void setDisplayLo (double lo)
    {
	displayLo = lo;
	commonAxis = true;
	rangeStyle = RANGE_STATIC;
	offGraphics = null;
    }

    /**
     * Return display high
     */
    public double getDisplayHi () { return displayHi; }


    /**
     * Return display low
     */
    public double getDisplayLo () { return displayLo; }


    /**
     * Start a new daemon thread to count the update time
     */
    private void start()
    {
        if (timer == null) { 
            timer = new Thread(this);
	    timer.setDaemon(true);
            timer.start();
        }
    }

    /**
     * Sleep a updatePeriod and call repaint method to update the chart
     */
    public void run() {
	while (true) {
	    try {
		Thread.sleep(updatePeriod);
	    }
	    catch (InterruptedException ie) {
		System.out.println( "thread failed to sleep");
	    }
	    repaint();
	}
    }

    /**
     * Set up offscreen graphics.  This is needed for both 
     * a new screen paint and a resize paint, or anything which
     * changes the widget's background (color, font, etc.).
     */
    protected void setupOffscreen() {
        offDimension = getSize();
        //System.out.println ("set new offDim size["+offDimension+"]...");
        offImage = createImage(offDimension.width, offDimension.height);
        offGraphics = offImage.getGraphics();
        offGraphics.setColor( bgColor );
        offGraphics.fillRect(0, 0, offDimension.width, offDimension.height);
    }

    /**
     * Overwrite the parent update method so it only call this.paint(g)
     */
    public final void update(Graphics g)
    {
        paint(g);
    }  

    /**
     * Draw offscreen graphics and repaint it onto the screen.
     * Calls derived class' paintForeground(g) and paintBackground(g).
     */
    public final void paint(Graphics g)
    {
	if (isShowing()) {
	    if (offGraphics == null) {
		setupOffscreen();
		paintBackground (offGraphics);
		if (history[0] != null && history[0].size() >1) 
		    drawHistory(offGraphics);
	    }
	    if (offGraphics != null) {
		paintForeground (offGraphics);
		g.drawImage(offImage, 0, 0, null);
	    }
	}
    }  


    private void drawHistory(Graphics g) {
	g.setClip(graphX, graphY, graphW, graphH);
	for (int j=0; j<penNum; j++) {
	    int cnt = history[j].size();
	    int x = graphX + graphW - (cnt-1)*dx;	
	    g.setColor(penColor[j]);
	    for (int i=0; i<cnt-1; i++) {
		g.drawLine(x, historyY(j, i), x+dx, historyY(j, i+1));
		x += dx;
	    }
	} 
    }


    private int historyY (int penNum, int pos) {
	double p;
	double d = ((Double)(history[penNum].elementAt(pos))).doubleValue();
	if (commonAxis)
	    p = (d-displayLo)/(displayHi-displayLo);
	else 
	    p = (d - lo[penNum])/(hi[penNum] - lo[penNum]);
	return (graphY + graphH - (int)(graphH*p));
    }


    private void initLabel () {
	fontH = fm.getHeight();
	// --- get pretty y axis ---------------------
	if (commonAxis) {
	    prettyY[0] = CUtility.makePrettyAxis(displayLo, displayHi);
	    displayLo = prettyY[0][0];
	    displayHi = prettyY[0][1];
	    stringYLo = CUtility.format(displayLo, fmt);
	    stringYHi = CUtility.format(displayHi, fmt);
	}
	else {
	    for (int i=0; i< penNum; i++) {
		prettyY[i] = CUtility.makePrettyAxis(lo[i], hi[i]);
		lo[i] = prettyY[i][0];
		hi[i] = prettyY[i][1];
	    }
	    stringYLo = CUtility.format(lo[0], fmt);
	    stringYHi = CUtility.format(hi[0], fmt);
	}

	//--- get pretty time axis (X axis) -------------
	if (displayPeriod <= 120) {        // use seconds
	    prettyX = CUtility.makePrettyAxis (0-displayPeriod, 0);
	    displayPeriod = 0 - (int)prettyX[0];
	    stringXLabel = "<--- Time (seconds)";
	}
	else  {                          // use minutes
	    prettyX = CUtility.makePrettyAxis(0-displayPeriod/60, 0);
	    displayPeriod = 0 - (int)prettyX[0]*60;
	    stringXLabel = "<--- Time (minutes)";
	}
    }

    private void initGraphics ()  {
	findRange();
	if (labelStyle != LABEL_NONE) initLabel();
	Dimension rec = getSize();
	faceW = rec.width;
	faceH = rec.height;
	gapX = (int) (0.04*faceW);
	gapY = (int) (0.04*faceH);

	//--- no break here to reduce the code -----
	switch (labelStyle) {
	case LABEL_ALL:
	    chaX = faceX + gapX;
	    chaY = faceY + gapY;
	    chaW = faceW - 2*gapX;
	    chaH = fontH + GAP;
	case LABEL_VALUE:
	case LABEL_AXIS:
	    xAxisH = fontH + TICK_LEN + GAP + GAP;
	    yAxisX = faceX + gapX;
	    yAxisY = faceY + gapY + chaH;
	    yAxisW = Math.max(fm.stringWidth(stringYLo),
			      fm.stringWidth(stringYHi))+2*TICK_LEN+GAP;
	    yAxisH = faceH - 2*gapY - chaH - xAxisH;
	    xAxisW = faceW - 2*gapX - yAxisW;
	    xAxisX = yAxisX + yAxisW;
	    xAxisY = faceH - gapY - xAxisH;
	default:
	    graphX = faceX + gapX + yAxisW;
	    graphY = faceY + gapY + chaH;
	    graphW = faceW - 2*gapX - yAxisW;
	    graphH = faceH - 2*gapY - xAxisH - chaH;
	}

	// --- find out the shift width dx ----------------
	dx = DEFAULT_DX;
	updatePeriod = (int)(1000*displayPeriod*dx/graphW);
	if (updatePeriod < MIN_UPDATE_PERIOD) {
 	    dx = Math.round((graphW*MIN_UPDATE_PERIOD)/(displayPeriod*1000));
	    updatePeriod = (int)(1000*dx*displayPeriod/graphW);
	}
	//System.out.println("dx = "+dx+",  update period="+updatePeriod+
	//         	   ",  display period="+displayPeriod);
	oldX = graphX + graphW - dx; 
	newX = graphX + graphW; 
	totalXPoint = (int) (graphW/dx);
					   
	for (int i=0; i< penNum; i++) {
	    oldY[i] = -1;
	    newY[i] = -1;
	    if (history[i] == null) history[i] = new Vector(totalXPoint);
	} 
    }

    private void drawYAxis(Graphics g) 
    {
        //--- draw label of Y axis ------------
        g.setColor(penColor[0]);
        g.setFont(font);
	strHiX = yAxisX;
	strHiY = yAxisY + fontH;
	strLoX = strHiX;
	strLoY = yAxisY + yAxisH; 
        g.drawString(stringYLo, strLoX, strLoY );
        g.drawString(stringYHi, strHiX, strHiY );

        //--- draw Y axis ticks -----------------
	int sY, eY;
	int sX = yAxisX + yAxisW;  
	int eX = sX - TICK_LEN;
	g.setColor(fgColor);
	for (int i = 0; i <= prettyY[0][2]; i++) {
	    double yinc = ((double)yAxisH)/prettyY[0][2];
	    sY = yAxisY + (int)(i*yinc);
	    eY = sY;
	    if (i==0 || i==(int)(0.5*prettyY[0][2]) || i==prettyY[0][2])   
		g.drawLine (sX, sY, eX-TICK_LEN, eY);
	    else 
		g.drawLine (sX, sY, eX, eY);
	}
    }

    private void findRange() 
    {
	for (int i=0; i< penNum; i++) {
	    if (!pen[i].gotRange())  offGraphics = null;
	    newValue[i] = pen[i].getNewValue();
	    hi[i] = pen[i].getDisplayHi();
	    lo[i] = pen[i].getDisplayLo();
	    penColor[i] = pen[i].getPenColor();
	    if (penColor[i] == null)  penColor[i] = defColor[i];
	}
	if (offGraphics != null) {
	    switch (rangeStyle) {
	    case RANGE_AUTO:
		commonAxis = true;
		displayHi = CUtility.arrayMax(newValue);
		displayLo = CUtility.arrayMin(newValue);
		System.out.println("hi="+displayHi+", low="+displayLo);
		break;
	    case RANGE_STATIC:	    
		// if (commonAxis) nothing need to do here
		// display range are set by user or use default values
		break;
	    default:        //case RANGE_FROM_PROPERTY:
		if (commonAxis) {
		    displayHi = CUtility.arrayMax(hi);
		    displayLo = CUtility.arrayMin(lo);
		}
		break;
	    }
	}
    }

    private void drawXAxis(Graphics g) 
    {
	//--- draw label of axis -------------
        int loX, hiX, labelX, y;
	String stringXLo, stringXHi = "0";
	stringXLo = CUtility.format(prettyX[0], "%.3g");
	stringXHi = CUtility.format(prettyX[1], "%.3g");
        g.setColor(fgColor);
        g.setFont(font);       
	loX = xAxisX - GAP - TICK_LEN;
	y = xAxisY + TICK_LEN + fontH + GAP;
	hiX = xAxisX + xAxisW - fm.stringWidth(stringXHi);
	labelX = xAxisX+(int)((xAxisW-fm.stringWidth(stringXLabel))/2); 
        g.drawString(stringXLo, loX, y);
        g.drawString(stringXHi, hiX, y);
        g.drawString(stringXLabel, labelX, y+GAP );

        //--- draw bar ticks -------------------------          
	int sX, eX;
	int sY = xAxisY;
	int eY = sY + TICK_LEN;
	for (int i=0; i<=prettyX[2]; i++) {
	    double xinc = ((double)xAxisW)/prettyX[2];
	    sX = xAxisX + (int)((double)i*xinc);
	    eX = sX;
	    if (i == 0 || i == (int)(0.5*prettyX[2]) || i == prettyX[2])   
		g.drawLine (sX, sY, eX, eY+TICK_LEN);
	    else                           
		g.drawLine (sX, sY, eX, eY);
	}
    }


    private void redrawBackground (Graphics g) {
	//--- erase old y axis label -----------------------
	g.setColor(bgColor);
	g.setClip(yAxisX, yAxisY, yAxisW-2*TICK_LEN, yAxisH);
	g.fillRect(yAxisX, yAxisY, yAxisW-2*TICK_LEN, yAxisH);
	//--- draw new y axis label -------------------------
	if (!commonAxis) {
	    stringYLo = CUtility.format(lo[currPen], fmt);
	    stringYHi = CUtility.format(hi[currPen], fmt);
	}
	g.setColor(penColor[currPen]);
	g.drawString(stringYLo, strLoX, strLoY);
	g.drawString(stringYHi, strHiX, strHiY);
	if (labelStyle == LABEL_ALL)  drawName(g);
	redrawBg = false;
    }

 
    protected void paintBackground (Graphics g) {
	initGraphics();
	g.setColor(bgColor);
	g.fill3DRect(faceX, faceY, faceW, faceH, true);

	switch (labelStyle) {
	case LABEL_ALL:
	    drawName(g);
	case LABEL_VALUE:
	case LABEL_AXIS:
	    drawYAxis(g);
	    drawXAxis(g);
	default:    //case LABEL_NONE:
	    g.setColor(fgColor);
	    g.drawRect(graphX, graphY, graphW, graphH);
	}
    }

    protected void paintForeground (Graphics g)
    {
	//--- redraw y axis if needed -------------------
	if ((labelStyle != LABEL_NONE) && (redrawBg)) 
	    redrawBackground(g);
	
	//--- shift the graph to left by dx and earse the dx slot -----
	g.copyArea(graphX+dx+1, graphY+1, graphW-dx, graphH-1, -dx, 0);
	g.setColor(bgColor);
        g.fillRect(oldX, graphY+1, dx, graphH-1);
       
	//--- prepare for new draw ------------------------
	for (int i=0; i< penNum; i++) {
	    oldY[i] = newY[i];
	    value = pen[i].getNewValue();
	    //--- find the auto display range ------------------
	    if (rangeStyle == RANGE_AUTO) {
		if (value > displayHi) { 
		    displayHi = value;
		    offGraphics = null;
		}
		if (value < displayLo) {
		    displayLo = value;
		    offGraphics = null;
		}
	    }
	    //--- calculate the y point for drawing --------------
	    if (commonAxis)
		percent = (value - displayLo)/(displayHi-displayLo);
	    else 
		percent = (value - lo[i])/(hi[i]-lo[i]);
	    newY[i] = graphY + graphH - (int) (graphH*percent);

	    //--- plot last one point(erased by y axis) and new point
	    if ((oldY[i] != -1) && (newY[i] != -1)) {
		g.setColor(penColor[i]);
		g.setClip(graphX, graphY, graphW, graphH);
		int s = history[i].size();
		if (s >= 2) 
		    g.drawLine(oldX-dx, historyY(i, s-2), oldX, oldY[i]);
		g.drawLine(oldX, oldY[i], newX, newY[i]);
	    }
	    //--- save the history data -------------------
	    if (history[i].size() == totalXPoint)
		history[i].removeElementAt(0);
	    history[i].addElement(new Double(value));
	}
    }
    
    private void drawName(Graphics g) 
    {
        //--- Erase old name -----------------------
	if (redrawBg) g.setClip (chaX, chaY, chaW, chaH);
        g.setColor (bgColor);
        g.fillRect(chaX, chaY, chaW, chaH);

	//--- draw new name -------------------------
        String name = pen[currPen].getDeviceName()+"."+
	              pen[currPen].getPropertyName();
        int stringW = fm.stringWidth(name);
        int stringX = chaX + (int)((chaW - stringW)/2);
        int stringY = chaY + chaH - GAP;  
        g.setColor (penColor[currPen]);
        g.drawString (name, stringX, stringY);
    }


    /**
     *  For mouseListener interface
     */
    public void mouseEntered  (MouseEvent me) { }
    public void mouseExited   (MouseEvent me) { }
    public void mouseReleased (MouseEvent me) { }
    public void mousePressed  (MouseEvent me) { }

    /**
     * For mouseListener interface handle mouse click
     * If click on one of the drawing line, the y axis and channel name 
     * will show the value of that line. Otherwise nothing happens.
     */
    public void mouseClicked (MouseEvent me)
    {
	int w = 3, h = 3;
	int x = me.getX();
        int y = me.getY();
	int[] pixels = new int[w * h];
	Color tmpColor = fgColor;
	boolean found = false;
	PixelGrabber pg = new PixelGrabber(offImage, 
			    x-1, y-1, w, h, pixels, 0, w);
	try {
            pg.grabPixels();
        } catch (InterruptedException e) {
            System.err.println("interrupted waiting for pixels!");
        }
	for (int j = 0; j < h; j++) {
            for (int i = 0; i < w && !found; i++) {
		tmpColor = new Color(pixels[j*w+i]);
		if (!tmpColor.equals(bgColor)) 
		    found = true;
            }
	}
	if (found) {
	    currPen = findPenNumber(tmpColor);
	    redrawBg = true;
	    repaint();
	}
    }
    
    private int findPenNumber (Color c) {
	boolean found = false;
	int num = 0, i = 0;
	for (i = 0; i < penNum && !found; i++) {
	    if (c.equals(penColor[i])) 
		found = true;
	}
	if (found)  num = i-1;
	return num;
    }
    
    /**
     * Overwrite parent setSize method to force repaint the background.
     *
     * @param newWidth     width of the component
     * @param newHeight    height of the component
     */
    public void setSize (int newWidth, int newHeight)
    {
        int w = Math.max(newWidth, MINSIZE);
        int h = Math.max(newHeight, MINSIZE);
	super.setSize(w, h);
        offGraphics = null;  
	repaint();
    }
    
   /** 
     * Set the bounds of the canvas.
     * @param x    component X location
     * @param y    component Y location
     * @param w    component width
     * @param h    component height
     */
    public void setBounds(int x, int y, int w, int h)
    {
        super.setBounds(x, y, w, h);
        offGraphics = null;   //clear the offGraphics to force a resize
    }


     /**
     * Set text font for labeling and displaying values.
     * @param newFont    font for any text on the component
     */
    public void setFont (Font newFont)  {
        font = newFont;
        offGraphics = null;     // force a repaint of the background
    }

    /**
     * Return the font of the text on the component
     */
    public Font getFont () { return font;}
  
    /**
     * Set component background color
     * @param newBg    the background color of component
     */
    public void setBackground (Color newBg) {
        bgColor = newBg;
        offGraphics = null;     // force a repaint of the background
    }
 

   /**
     * Set compoenet foreground color
     * @param newFg   the foreground color of component
     */
    public void setForeground (Color newFg) { 
        fgColor = newFg;
        offGraphics = null;     // force a repaint of the background
    }

} // end CStripChart



