/*
 * Copyright (c) 2000 Southeastern Universities Research Association,
 *                    Continuous Electron Beam Accelerator Facility
 *
 * This software was developed under a United States Government license
 * described in the NOTICE file included as part of this distribution.
 *
 *
 *  Revision History:
 *  09/29/00:  Initial version.
 *
 */

package adm;

import java.awt.*;
import java.awt.event.*;

/**
 * This widget contains one CTextField on the left and two ArrowButtons
 * on the right. It only works when connect to the device-property with 
 * numerical number value.
 * <BR><BR>
 * The textfield is used to display the value and the two arrow buttons
 * are used to adjust the value. The digit at left of the cursor will 
 * increase or decrease one when user clicks the up arrow button or the 
 * down arrow button. 
 * <BR><BR>
 * Like CTextField, the device name, property name and color mode can
 * be controlled by the user. The color mode has the default value,
 * CConstants.CLRMOD_STATIC. But the device name and property name must 
 * be set when using this widget.
 * 
 *
 * @version   1.0  10/10/00
 * @author    Ying Chen
 *            
 * @since     adm1.0
 */
public class CWheelSwitch extends Container implements ActionListener 
{

    private double increment = 1.0;

    private CTextField textfield;
    private ArrowButton upB;
    private ArrowButton downB;
    
    /**
     * Null Constructor
     */
    public CWheelSwitch ()
    {
	initComponent();
    }

    public CWheelSwitch (String dname, String pname)
    {
	this();
	setDeviceName(dname);
	setPropertyName(pname);
    }

    private void initComponent() 
    {
	textfield = new CTextField();
	textfield.setFocusLock(false);
	textfield.setFormat("%.3f");

	upB = new ArrowButton(CConstants.DIR_UP);
	upB.addActionListener(this);
	downB = new ArrowButton(CConstants.DIR_DOWN);
	downB.addActionListener(this);
	Panel arrowP = new Panel();
	arrowP.setLayout(new GridLayout(2, 1, 0, 2));
	arrowP.add(upB);
	arrowP.add(downB);

	setLayout(new BorderLayout());
	add(BorderLayout.CENTER, textfield);
	add(BorderLayout.EAST, arrowP);
    }
    
    public void actionPerformed (ActionEvent e) {
	int cPos = textfield.getCaretPosition();
	ArrowButton ab = (ArrowButton) e.getSource();
	if (ab == downB || ab == upB) {
	    String valString = textfield.getText();
	    int dPos = valString.indexOf(".");
	    if (cPos > dPos)
		increment = Math.pow(10, dPos - cPos + 1);
	    else {
		if (cPos == 0) cPos = valString.length();
		textfield.setCaretPosition(cPos);
		increment = Math.pow(10, dPos - cPos);
	    }
	    try {
		double val = Double.valueOf(valString).doubleValue();
		if (ab == upB)   val = val + increment;
		else             val = val - increment;
		//System.out.println("after val ="+val);
		textfield.setText(String.valueOf(val));
	    } catch (Exception ee) {
		System.out.print("ERROR: value is not a numerical number.");
	    }
	}
    }


    /**
     * Set the Device name 
     * @param newName     the name of device you want to connect
     */
    synchronized public void setDeviceName(String newName)
    {
	textfield.setDeviceName(newName);
    }
    

   /**
     * Return the device name.
     * If device name is not set, return empty string "". 
     */
    public String getDeviceName()
    {
	return textfield.getDeviceName();
    }
    

    /**
     * Set device property name. 
     * @param newName    the name of device property you want to connect
     */  
    synchronized public void setPropertyName(String newName)
    {
	textfield.setPropertyName(newName);
    }


     /**
     * Return the device property name.
     * If property name is not set, return empty string "". 
     */
    public String getPropertyName()
    {
	return textfield.getPropertyName();
    }
    
    public void setColorMode (int mode) 
    {
	textfield.setColorMode(mode);
    }

    public int getColorMode () 
    {
	return textfield.getColorMode();
    }
}
