//---------------------------------------------------------------------------
// Copyright (c) 1996,1997 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      Java implementation of a class which routes I/O requests to a service.
//      This class manages transactions, and delivers return events
//      as responses arrive from the IOservice. 
//
//  Authors: Cuong Quach (Patrick) and Chip Watson
//
//  Partially based on work by Walt Akers and Jie Chen, plus ideas from 
//  Franck DiMaio at CERN
//                                       
//  ****NOTE: pend() is improperly implemented -- should only wait if I/O is pending
//
//----------------------------------------------------------------------------
// Revision History:
//
// 10-3-96 : Created by CQ.
//           ...intervening history deleted
// 8-11-98:  CW Modified to prototype proposed version 2 changes. Some functionality
//           moved to CLIPservice class, used via IOservice interface.
// 9-3-98:   FDM checked the prototype against a PS JNI service.
//           Add usage of a Config object for obtaining IOService(s).
//           Modifications in  DEFAULT_PEND_PERIOD, handleResponse and findTrans.
//           Suppressed one access to package-visible attributes of Data objects.
// 9-8-98:   CW replaced Fifo with IntFifo, 
//           changed handleResponse arguments to eliminate DeviceResponse class
// 10-5-98:  Fixed bugs related to outbound data; changed tr.status to tr.state
// 11-19-98: Added public findTrans(int trnum); make handleResponse static
//  5-Dec-98: CW Change post to directly use transaction + related changes
//  7-Dec-98: CW Added throws clauses to I/O calls
// 18-Jan-99: CW Fixed bug in occupancy tracking created when activeCnt added
// 12-Apr-99: MC changed import for IOService now located in cdev.io
// 23-Apr-99: CW Split out TransactionManager class
// 28-May-99: CW replaced IntFifo with Fifo (race condition danger)
// 22-Oct-99: MC modified DEFAULT_PEND_PERIOD-> 5s, 
//	         add a System Property for pend(dt) timeOut
//               Timeout is now first taken from DeviceContext if existing
// 25-Oct-99: MC added removeTrans(dt) for deleting transactions from package level 
//		 (used in Device.pend(dt))
// 15-Nov-99: CW added suspendIO, resumeIO for fast applet exit and re-enter
// 10-Apr-00: CW optimize ordering within handleResponse for better performance
//               and make buffering of events and deliver in ControlSystem thread
//               optional, turned off by default for better performance
//
//----------------------------------------------------------------------------

package cdev;

import cdev.data.*;
import cdev.io.*;
import cdev.dir.*;

import java.io.*;
import java.net.*;
import java.util.*;

/**
 *
 * @version 2-1 beta 2
 * <br>
 * Java implementation of the class which manages I/O
*/
public class ControlSystem implements Runnable
{
  //-----------------------------------------------
  // Return codes for IO methods:
  //-----------------------------------------------

  /**
   * Failure is non-consequential.
  */
  public static final byte CDEV_WARNING       =  -2;  

  /**
   * Errors that are not in any categories.
   */
  public static final byte CDEV_ERROR         =  -1;

  /**
   * CDEV success.
   */
  public static final byte CDEV_SUCCESS       =   0; 

  /**
   *  Invalid CDEV objects.
   */
  public static final byte CDEV_INVALIDOBJ    =   1; 

  /**
   *  Invalid argument passed to cdev calls
   */
  public static final byte CDEV_INVALIDARG    =   2; 

  /**
   * Wrong service during dynamic loading.
   */
  public static final byte CDEV_INVALIDSVC    =   3; 

  /**
   * Operation is unsupported.
   */
  public static final byte CDEV_INVALIDOP     =   4; 

  /**
   * CSystem not connected to low network service.
   */
  public static final byte CDEV_NOTCONNECTED  =   5; 

  /**
   * Low level network service IO failed.
   */
  public static final byte CDEV_IOFAILED      =   6; 

  /**
   * Conflicts of data types or tags. 
   */
  public static final byte CDEV_CONFLICT      =   7; 

  /**
   * CDEV cannot find user request.
   */
  public static final byte CDEV_NOTFOUND      =   8; 

  /**
   * Time out.
   */
  public static final byte CDEV_TIMEOUT       =   9; 

  /**
   * cdev.Data conversion error.
   */
  public static final byte CDEV_CONVERT       =  10; 

  /**
   * Value out of range for device attribute.
   */
  public static final byte CDEV_OUTOFRANGE    =  11; 

  /**
   * Insufficient access to perform request.
   */
  public static final byte CDEV_NOACCESS      =  12; 

  /**
   * Change in access permission of device.
   */
  public static final byte CDEV_ACCESSCHANGED =  13; 

  /**
   * Channel has been disconnected.
   */
  public static final byte CDEV_DISCONNECTED     =   60;

  /**
   * Channel has been reconnected.
   */
  public static final byte CDEV_RECONNECTED      =   61;

  /**
   * The callback object will be deleted
   */
  public static final byte CDEV_DELETE_CALLBACK  =   70;

  /**
   * More data to come.
   */
  public static final byte CDEV_INCOMPLETE      = 88; 

  /** pend period in milliseconds. */
  static final long DEFAULT_PEND_PERIOD = 5000; // 5 seconds
  static long pendPeriod = DEFAULT_PEND_PERIOD; // default for pending


  //-----------------------------------------------------------------//
  // controlSystemTable: used to locate ControlSystem objects by name
  //-----------------------------------------------------------------//
  private static final int   HASHTABLESIZE = 20;
  private static final float HASHLOAD = (float) 0.25;
  private static Hashtable controlSystemTable = new Hashtable(HASHTABLESIZE,HASHLOAD);
  private static TransactionManager tmgr = new TransactionManager();
  private static Vector services = new Vector();

  //-------------------------------------------------------------//
  // fifo:
  // Private data for queue of unprocessed transaction responses
  //-------------------------------------------------------------//
  private static final int DEFAULT_FIFO_CAPACITY = 500;
  private Fifo fifo = null;     // ring buffer of transactions with results
 
  Thread thread = null;
  static final int PRIORITY_BOOST = 1; // boost ControlSystem's thread priority
  static final int BUFFER_BEFORE_YIELD = 0; // when handling responses
  boolean buffered = false;	// false==deliver events in IOService thread

  // Other instance data
  String name = null;
  boolean active = false;

  //**********************************************************************************//
  /**
   * Constructor.
   */
  public ControlSystem () throws IOException
  {
    name = getDefaultName();
    init();
  }

  //**********************************************************************************//
  /**
   * Construct a ControlSystem object with a given name.
   * @param  name   The name of the ControlSystem
   */
  public ControlSystem (String name) throws IOException
  {
    this.name = name;
    init();
  }

  public String getName() {return name;}

  private void init() throws IOException {
    Object test = controlSystemTable.get(name);
    if (test != null) 
      throw new IOException("Duplicate ControlSystem creation attempted");
    tmgr = new TransactionManager();
    controlSystemTable.put(name,this);
    fifo = new Fifo (DEFAULT_FIFO_CAPACITY);
    thread = new Thread(this);
    thread.setName (name); // set name of this Thread
    try {thread.setDaemon(true);} catch (Exception e) {}
    try {
      thread.setPriority(thread.getPriority() + PRIORITY_BOOST);
    } catch (Exception e) {}
    thread.start();
    active = true;
  }

  //**********************************************************************************//
  /**
   * Get the default ControlSystem object for a given name.
   * @param  name   The name of the ControlSystem.
   */
  public static ControlSystem getDefault (String name) throws IOException
  {
    ControlSystem sys = (ControlSystem) controlSystemTable.get(name);
    if (sys == null) sys = new ControlSystem(name);
    return sys;
  }

  //**********************************************************************************//
  /**
   * Get the default ControlSystem object.
   */
  public static ControlSystem getDefault() throws IOException
  {
    return getDefault(getDefaultName());
  }

  // helper routine to get the default control system name,
  // currently the name of the current thread group
  static String getDefaultName() 
  {
    return Thread.currentThread().getThreadGroup().getName();
  }

  //**********************************************************************************//
  /**
   * Register an IOService with this ControlSystem
   */
  public void registerIOService(IOService s) {
    services.removeElement(s);
    services.addElement(s);
  }
  public void unregisterIOService(IOService s) {
    services.removeElement(s);
  }

  //**********************************************************************************//
  /**
   * Flushes any pending I/O to the server(s).
   */
  public void flush () throws IOException
  {
    for (int i=0;i<services.size();i++) {
      IOService service = (IOService) services.elementAt(i);
      service.flush();
    }
  }

  // Convenience routines for connection management, for single 
  // service operation (defer to service, if one exists)

  public void connect (String host, int port) throws Exception {
    IOService service = getSingleService();
    service.connect(host,port);
  }
  public void disconnect () throws IOException {
    if (services.size() == 0) return;
    IOService service = getSingleService();
    service.disconnect();
  }
  public void reconnect () throws Exception {
    IOService service = getSingleService();
    service.reconnect();
  }
  public boolean isConnected() throws IOException {
    if (services.size() == 0) return false;
    IOService service = getSingleService();
    return service.isConnected();
  }
  public String getHostname () throws IOException {
    return getHostName();
  }
  public String getHostName () throws IOException { 
    IOService service = getSingleService();
    return service.getHostName();
  }
  public int getPort() throws IOException { 
    IOService service = getSingleService();
    return service.getPort();
  }
  IOService getSingleService() throws IOException {
    IOService service = null;
    int count = services.size();
    if (count==1)
      service = (IOService) services.firstElement();
    else if (count==0) {
      try {
	service = Config.getIOService(null);
      } catch (Exception e) {
	throw new IOException("Failed to initialize I/O service "+e);
      }
      if (service != null) services.addElement(service);
    }
    if (service == null)
      throw new IOException("Can not get unique I/O service");
    return service;
  }

  //**********************************************************************************//
  /**
   * Device factory method. Currently just invokes device constructor.
   * Provides backward compatibility with CDEV 1.x, plus future options.
   * @param  name   Name of the device to get.
   */
  public Device getDevice (String name) throws IOException
  {
    return new Device(name,this);
  }

  //**********************************************************************************//
  /**
   * Wait for server responses for the default duration.  Note that
   * events may be delivered before a pend() is called.
   */
  public void pend () throws IOException
  {
    pend (pendPeriod);
  }

  /**
   * Wait for server responses for the given duration.  If "millisec" is 
   * 0, pend will NEVER return.
   * @param  sec        The pend period in seconds.  If < 0.001, then pend forever!
   */
  public void pend (double sec) throws IOException
  {
    pend ((long)(sec*1000.0));
  }

  /**
   * Wait for server responses for the given duration.  If "millisec" is 
   * 0, pend will NEVER return.
   * @param  millisec        The pend period in milliseconds.  If 0, then pend forever!
   */
  public void pend (long millisec) throws IOException
  {
    flush();

    // the following sleeps in 10 ms increments; a future
    // version should be fully async (signal)
    if (millisec == 0) {
      while (tmgr.getActiveCnt() > 0) {
	  try {
	      Thread.sleep(10); // sleep in 10 millisecond intervals  
	  } catch (Exception e){}
      }
    } else {
      long endtime = millisec + System.currentTimeMillis();
      while (System.currentTimeMillis() < endtime && tmgr.getActiveCnt() > 0)  {
	long timeleft = endtime - System.currentTimeMillis();
	if (timeleft > 10) timeleft = 10;
	try {
	  Thread.sleep(timeleft);
	} catch (Exception e){}
      }
    }
  }

  //--------------------------------------------------------
  /**
   * Special pend used by Device for implementing synchronous transactions.
   * Timeout value is taken from the DeviceContext if present; otherwise,
   * it is taken from a System Property(cdev.pendTimeOut) as a long value. 
   * If user has not set the Property, default is DEFAULT_PEND_PERIOD.
   * @param  dt         Reference to the device transaction to use
   *                    for synchronization.
   */
  //--------------------------------------------------------
  void pend (DeviceTransaction dt) throws IOException
  {
    flush();
    synchronized (dt) {
      if (dt.state == DeviceTransaction.PENDING || 
	  dt.state == DeviceTransaction.INCOMPLETE) {
	DataEntry timeOutEntry;
	timeOutEntry = dt.getDevice().getContext().getData().getDataEntry("timeOut");
	long waitTime = (timeOutEntry==null)? pendPeriod: timeOutEntry.longValue();
	try { dt.wait(waitTime); } catch (InterruptedException e) {}
      }
    }
  }
    
  //**********************************************************************************//
  /**
    * post:
    * Post a transaction into the system transaction table
    */
  final void post ( DeviceTransaction dt ) throws IOException {
    if (active) {
      tmgr.add(dt);
      IOService service = dt.getDevice().getIOService();
      service.post(dt);
    }
  }

  //**********************************************************************************//
  /**
    * postMonitorOn:
    * Post a MonitorOn transaction into the system transaction table,
    * compressing with any duplicate transaction (use same results when they arrive).
    * The last argument is an Object so that this method can handle sendNoBlock calls.
    * The transactionHandler will use the instanceof operator to determine what
    * this is and carry out the appropriate action.  If it is an instanceof Data
    * the handler only has to copy the result over.  If it is an instanceof 
    * DeviceListener, then the transactionHandler will call the appropriate method
    * of that object.
    */
  final void postMonitorOn ( Device device, String property, DeviceListener listener ) 
       throws IOException {

    if (active) {
      if (!tmgr.addListener(device.getName(), property,
			    device.getContext().getSequenceNumber(), listener)) {
	DeviceTransaction dt = new DeviceTransaction(device,DeviceTransaction.MONITORON,
						     property,null,listener);
	tmgr.add(dt);
	IOService service = device.getIOService();
	service.post(dt);
      }
    }
  }

  //**********************************************************************************//
  /**
    * postMonitorOff:
    * Look for a matching "monitorOn" and remove it
    * if there is only one pending listener. 
    * Otherwise just remove this listener.  If no
    * matching transactions are found, ignore this call.
    */
  final void postMonitorOff (Device device, String property, Object listener ) 
       throws IOException {

    if (active) {
      DeviceTransaction dt;
      dt = tmgr.removeListener(device.getName(), property,
			       device.getContext().getSequenceNumber(), listener);
      if (dt != null) {
	IOService service = dt.getDevice().getIOService();
	service.cancel(dt);
      }
    }
  }

  //*****************************************************************************
  /**
   *  findTransaction:
   *  Called by IOService to find a pending transaction based upon the
   *  transaction number (used for bookkeeping).
   */
  public static final DeviceTransaction findTransaction(int trnum) {
    return tmgr.findTransaction(trnum);
  }

  //*******************************************************************************//
  /**
   *  handleResponse:
   *  Called by IOService to deliver results from an earlier transaction posting.
   *  Prior to this call, the IOService must update the data, error, and state
   *  fields of the transaction.
   *  @param dt    DeviceTransaction
   */
  public static void handleResponse(int trnum, int status, Data results, DeviceError err)
  {
    DeviceTransaction dt = tmgr.findTransaction(trnum);    
    //System.out.println("handleResponse "+trnum+": "+dt);
    if (dt != null) {
      if (dt.state != DeviceTransaction.COMPLETE) { // false for most monitors
	if (status == CDEV_INCOMPLETE) 
	  dt.state = DeviceTransaction.INCOMPLETE;
	else
	  dt.state = DeviceTransaction.RECEIVED;
      }

      dt.error = err;

      // dt.data may be
      //   null (no data received yet)
      //   users Data object (sync get, sync send, or sendNoblock)
      //   transient data object (left from last update of this transaction)
      //   persistent data object (perhaps shared with other transactions)
      //   

      // deal with synchronous case first
      if (dt.listener == null) {
	if (dt.data != null && dt.data != results && results != null) {
	  // copy results to user's data object
	  try {
	    results.copyInto(dt.data); // probably user's data object
	  } catch (CloneNotSupportedException e) {
	    dt.data.remove(); // clear the Data
	    err = new DeviceError(dt,CDEV_INVALIDOBJ, //add an error to posted result
				  "Error in ControlSystem while returning data"); 
	  }
	}
	// process immediately since there is no user code to worry about
	tmgr.process(dt);
      }

      // asynchronous case (one or more listeners)
      else {
	dt.data = results; // may already be same if service is re-using
	ControlSystem cs = dt.device.getControlSystem();

	// notify controlSystem's thread if in buffered mode
	if (cs.buffered) {
	  cs.add(dt); // defer invoking listener to system's thread
	  synchronized (cs) {
	    cs.notify();  // Notify ControlSystem's waiting thread
	  }
	  if (cs.fifo.size() > BUFFER_BEFORE_YIELD) Thread.yield();
	} 

	// Process immediately if not in not buffered mode
	else {
	  tmgr.process(dt);
	}
      }
    }
  }

  private void add(DeviceTransaction dt) {
    if (active) fifo.add(dt);
  }

  //---------------------------------------------------------------------
  // run:
  // Handle response events from the service(s)/server(s).
  // Dispatch events to listeners.  If just a result data is required, copy
  // the data and wake any waiting threads.  Delete the transaction when 
  // done handling the response.
  //---------------------------------------------------------------------
  public void run () {

    while (active) {    // go forever

      //--- wait to be notified by the service ---
      synchronized (this) {
        if (fifo.isEmpty())
	  try { this.wait(); } catch (Exception e) {}
      }
      //--- process all transactions in the fifo ---
      if (active) tmgr.process (fifo);

      //tmgr.process(dt); // pass to TransactionManager for processing
    }
    //System.out.println("handle done.");
  }

  /**
   * Suspend I/O
   */
  public void suspendIO () throws IOException {
    //System.out.println("suspendIO "+active);
    if (active) {
      tmgr.suspendIO(this); // cancel all monitors for this system
      flush();
      try {Thread.currentThread().sleep(10);} catch (Exception e) {}
      disconnect(); // drop connections (stops IOService threads)
      active = false; // disable new transactions
      synchronized (this) {this.notify();}
      Thread.yield();
      thread = null;
    }
  }

  /**
   * Resume I/O
   */
  public void resumeIO() throws Exception {
    //System.out.println("resumeIO "+active);
    if (!active) {
      thread = new Thread(this);
      active = true;
      thread.start();
      reconnect();
      try {
	  Thread.sleep(10); // make sure other threads can run
      } catch (Exception e) {}
      tmgr.resumeIO(this);
    }
  }

  /**
   * Finish all use of this ControlSystem. Use of this object after 
   * calling finish() is undefined. 
   */
  public void finish() {
    controlSystemTable.remove(name);
    try {
      suspendIO();
    } catch (Exception e) {}
    name = null;
    fifo = null;
  }

  //*******************************************************************//
  /**
   * Stop transactions associated with this ControlSystem, and
   * free up memory before destroying the ControlSystem object.
   */
  protected void finalize () throws Throwable
  {
    finish();
    super.finalize();
  }


  // *******************************************************************************
  //-------------------------------------------------------------
  // Debug methods:
  //-------------------------------------------------------------

  /**
   * Convert the transaction table to a String.
   * The Title is prepended to the result.
   */
  public String txTableToString (String title)
  {
    return tmgr.txTableToString(title);
  }

  //----------------------------------------------------
  // static class initializer
  //----------------------------------------------------

  static {
    // trying to get default pending timeout for pend(dt) from a System Property
    // default is in DEFAULT_PEND_PERIOD declaration
    String pendTimeOut = null;
    try {
      pendTimeOut = System.getProperty("cdev.pendTimeOut");
    }
    catch(Exception e) {
      //System.err.println("cdev.pendTimeOut System Property"+e.toString()+
      //"...taking "+DEFAULT_PEND_PERIOD+"s as default");
    }
    if (pendTimeOut != null) {
      pendPeriod = Long.valueOf(pendTimeOut).longValue();
    }
  }
}   // end of ControlSystem class


