//---------------------------------------------------------------------------
// Copyright (c) 1991,1992 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      Java implementation of cdev Device abstraction for control system I/O.
//
//  Authors: Cuong Quach (Patrick) and Chip Watson
//
//  Partially based on work by Walt Akers and Jie Chen.
//                                       
//----------------------------------------------------------------------------
// Revision History:
//
//  9-26-96 : Created CQ.
// 11-21-96 : Removed "unregOn" variable and "finalize" method.
//            Implement "send" using "sendNoBlock".
//  2-10-97:  Remove cdev prefix from class name references.
//  4-13-98:  Changed send() method to pend until data is returned.  Previously,
//            it would pend() until the earlier of the default pend period of the
//            reply data is returned from the server.    cq
//  4-20-98:  Add predefined static contexts which may be used with messages.
//  4-29-98:  Reinstate the old behavior for send() to return the sooner of
//               1) a time out after default wait period  or
//               2) the server replies.
//  8-11-98:  CW: prototype proposed changes for version 2.
//  9-3-98:   FDM: checked the prototype against a PS JNI service.
//            2 access methods made public.
//  9-24-98:  CW: added back in the deprecated sendCallback() method
//  11-17-98: FDM added directory service methods
//  11-23-98: FDM added reference I/O methods
//  11-30-98: FDM added IO service data & access method.
//  8-Jan-99: CW added throws clause to all I/O methods
//  27-Apr-99: M.C added sync and async getCharacteristic.
//  27-Apr-99: M.C added test(and throw) on parameter Data in get methods.
//   3-Jun-99: CW added wrapper class to handle deprecated Callback methods
//   3-Jun-99: CW made directory service optional (return null)
//  16-Jul-99: FDM re-added exception for implicit directory service calls (proposal) 
//
//----------------------------------------------------------------------------

package cdev;

import cdev.data.*;
import cdev.dir.*;
import cdev.io.*;

import java.io.*;
import java.util.*;

/**
 * This class represents the device abstraction in the cdev Java paradigm. 
 * Devices have properties which can be set or retrieved via synchronous or
 * asynchronous methods (determined by argument type):
 * <UL>
 *    <LI> get(String property, ...) - get value of property
 *    <LI> set(String property, ...) - set value of property
 *    <LI> monitorOn(String property,...) - start monitoring value of property
 *    <LI> monitorOff(String property,...) - stop monitoring value of property
 * </UL>
 * Devices can also be sent messages using any of three methods.
 * <UL>
 *    <LI> send() - synchronous send.
 *    <LI> sendNoBlock() - asynchronous send.
 *    <LI> send(,,,listener) - asynchronous send with callback to listener
 * </UL>
 * Four basic messages are currently implemented for any property.  A
 * message may contain a keyword (verb) and one additional word naming
 * the property.  No limit is set on the number of words in a message for
 * non-property based I/O.  The keywords are the same as the property methods:
 * <UL>
 *    <LI> get
 *    <LI> set
 *    <LI> monitorOn
 *    <LI> monitorOff
 * </UL>
 * See the user documentation for more details about use of messages and
 * format.
 * @see cdev.data.Data
 * @see cdev.DeviceContext
 */
public class Device
{
  //------------------------------------------------
  // Define and initialize static context variables
  // for use with messages.
  //------------------------------------------------

  // Note: (CW) the following statics should be reviewed

  public static Data CONTROL_CTX  = new Data();
  public static Data MONITOR_CTX  = new Data();
  public static Data VALUE_CTX    = new Data();
  static {
      //--- initialize CONTROL_CTX
      CONTROL_CTX.insert("value", 1);
      CONTROL_CTX.insert("displayHigh",1);
      CONTROL_CTX.insert("displayLow",1);
      CONTROL_CTX.insert("alarmHigh",1);
      CONTROL_CTX.insert("alarmLow",1);
      CONTROL_CTX.insert("warningHigh",1);
      CONTROL_CTX.insert("warningLow",1);
      CONTROL_CTX.insert("controlHigh",1);
      CONTROL_CTX.insert("controlLow",1);

      //--- initialize MONITOR_CTX
      MONITOR_CTX.insert("value", 1);
      MONITOR_CTX.insert("displayHigh",1);
      MONITOR_CTX.insert("displayLow",1);
      MONITOR_CTX.insert("alarmHigh",1);
      MONITOR_CTX.insert("alarmLow",1);
      MONITOR_CTX.insert("warningHigh",1);
      MONITOR_CTX.insert("warningLow",1);

      //--- initialize VALUE_CTX
      VALUE_CTX.insert("value", 1);
      VALUE_CTX.insert("displayHigh",1);
      VALUE_CTX.insert("displayLow",1);
      VALUE_CTX.insert("alarmHigh",1);
      VALUE_CTX.insert("alarmLow",1);
      VALUE_CTX.insert("warningHigh",1);
      VALUE_CTX.insert("warningLow",1);
  }

  private ControlSystem parentSys;
  private String deviceName;
  private DeviceContext context;
  private IOService IOservice;
  private Object serviceData; // for use by IOservice; not used elsewhere in cdev

  //**************************************************************************
  // Constructor
  //**************************************************************************
  public Device ( String dname ) throws IOException 
  {
    deviceName = dname;
    parentSys = ControlSystem.getDefault();
    context = DeviceContext.getDefault();
  }

  public Device ( String dname, ControlSystem sys ) throws IOException
  {
    deviceName = dname;
    parentSys = sys;
    if (parentSys == null) {
      parentSys = ControlSystem.getDefault();
    }
    context = DeviceContext.getDefault();
  }

  //--------------------------------------------------------------//
  // get:
  /**
   * Initiate asynchronous get of property value. If the value is
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param listener   Reference to the device listener to receive value
   */
  //--------------------------------------------------------------//

  public void get (String property, DeviceListener listener) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.GET,property,null,listener);
    parentSys.post(dt);
  }

  //--------------------------------------------------------------//
  // get:
  /**
   * Initiate synchronous get of property value. If the value is
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param result     Reference to the data object to receive the result
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//

  public DeviceError get (String property, Data result) throws IOException {
    if (result == null) 
	throw new IllegalArgumentException("Data object parameter can't be null");
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.GET,property,null,result);
    parentSys.post(dt);
    return pend(dt);
  }

  //--------------------------------------------------------------//
  // set:
  /**
   * Initiate asynchronous set of property value. An acknowledgement
   * is delivered to the listener when a response is received. If a 
   * response is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param data       Data object containing the value.
   * @param listener   The device listener object to receive the response.
   */
  //--------------------------------------------------------------//

  public void set (String property, Data data, DeviceListener listener) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SET,property,data,listener);
    parentSys.post(dt);
  }

  //--------------------------------------------------------------//
  // set:
  /**
   * Initiate synchronous set of property value. If a response
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param data       Data object containing the value.
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//

  public DeviceError set (String property, Data data) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SET,property,data,
				(DeviceListener)null);
    parentSys.post(dt);
    return pend(dt);
  }

  // monitorOn:
  /**
   * Initiate asynchronous subscription to property value. If a value is
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param listener   Reference to the device listener to receive the results
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//

  public void monitorOn (String property, DeviceListener listener) throws IOException {
    parentSys.postMonitorOn(this,property,listener);
  }

  // monitorOff:
  /**
   * Terminate asynchronous subscription to property value.
   *
   * @param property   The name of the property
   * @param listener   Reference to the device listener to receive the results
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//

  public void monitorOff (String property, DeviceListener listener) throws IOException {
    parentSys.postMonitorOff(this,property,listener);
  }

  //-------------------------------------------------------------//
  // send:
  /**
   * Initiate a transaction that delivers results to a listener
   * when the reply is returned from the server.
   * @param message    The message to send to the server.
   * @param dataIn     The data to send with the message.
   * @param listener   The listener object to receive results.
   */
  //-------------------------------------------------------------//
  public void send ( String message, Data data, DeviceListener listener ) throws IOException {
    // convert send of monitorOn/Off to simpler form
    StringTokenizer msg = new StringTokenizer(message," ");
    String verb      = msg.nextToken();
    if (verb.equals("monitorOn"))
      monitorOn(msg.nextToken(),listener);
    else if (verb.equals("monitorOff"))
      monitorOff(msg.nextToken(),listener);
    else {
      DeviceTransaction dt;
      dt = new DeviceTransaction (this,DeviceTransaction.SEND,message,data,listener);
      parentSys.post(dt);
    }
  }

  //-------------------------------------------------------------//
  // send:
  /**
   * Initiate a synchronous "send" transaction.  This method
   * will return the sooner of receiving a reply or the default time
   * period expires.  When returning as a result of timeout, the result
   * will be unchanged from what was passed in.
   * @param message   The message to send to the server.
   * @param data      The data to send with the message.
   * @param result    Reference to data to return from the server.
   * @return          The status from the transaction. 
   */
  //-------------------------------------------------------------//
  public DeviceError send ( String  message
		    ,Data    data 
		    ,Data    result  ) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SEND,message,data,result);
    parentSys.post(dt);
    return pend(dt);
  }


  //-------------------------------------------------------------//
  // sendNoBlock:
  /**
   * Initiate asynchronous transaction that delivers data to Data object.
   * @param message   The message to send to the server.
   * @param data      The data to send with the message.
   * @param result    Reference to data to receive result from the server.
   */
  //-------------------------------------------------------------//
  public void sendNoBlock ( String  message
		   	   ,Data    data
		    	   ,Data    result ) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SEND,message,data,result);
    parentSys.post(dt);
  }

  // sendCallback (deprecated)
  /**
   * @deprecated Use send with DeviceListener instead.
   */
  public void send (String message, Data data, Callback handler) throws IOException {
    CallbackAdaptor ca = new CallbackAdaptor();
    ca.setCallback(handler);
    send(message,data,ca);
  }

  private class CallbackAdaptor implements DeviceListener {
    private Callback callback;
    public void setCallback(Callback c) {callback = c;}
    public void deviceChanged (DeviceEvent evt) {
      DeviceError err = evt.getError();
      int errcode = (err == null)? 0: err.getCode();
      callback.execute(errcode, evt.getDevice(), evt.getMessage(), evt.getData());
    }
  }

  //-------------------------------------------------------------
  // Device sync pend utility routine
  //

  private DeviceError pend (DeviceTransaction dt) throws IOException {
    parentSys.pend(dt);
    return dt.error;
  }

  //-------------------------------------------------------------//
  // getReference:
  /**
   * Initiate synchronous get of property reference value. 
   * If the value is not received within TIMEOUT seconds, 
   * an error is returned.
   * 
   * @param property   The name of the property
   * @param result     Reference to the data object to receive the result
   * @return           DeviceError reference, or null if no errors
   */
  public DeviceError getReference (String property, Data result) throws IOException {
    if (result == null) 
      throw new IllegalArgumentException("Data object parameter can't be null");
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.GETREFERENCE,property,null,result);
    parentSys.post(dt);
    return pend(dt);
  }
  /**
   * Initiate asynchronous get of property reference value. If the value is
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param listener   Reference to the device listener to receive value
   */
  //--------------------------------------------------------------//
  public void getReference (String property, DeviceListener listener) throws IOException{
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.GETREFERENCE,property,null,listener);
    parentSys.post(dt);
  }

  //--------------------------------------------------------------//
  // setReference:
  /**
   * Initiate asynchronous set of property reference value. An acknowledgement
   * is delivered to the listener when a response is received. If a 
   * response is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param data       Data object containing the value.
   * @param listener   The device listener object to receive the response.
   */
  //--------------------------------------------------------------//
  public void setReference (String property, Data data, DeviceListener listener) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SETREFERENCE,property,data,listener);
    parentSys.post(dt);
  }
  /**
   * Initiate synchronous set of property reference value. If a response
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param data       Data object containing the value.
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//
  public DeviceError setReference (String property, Data data) throws IOException {
    DeviceTransaction dt;
    dt = new DeviceTransaction (this,DeviceTransaction.SETREFERENCE,property,data,
				(DeviceListener)null);
    parentSys.post(dt);
    return pend(dt);
  }

  /**
   * Initiate synchronous get value of a property's characteristic . If a response
   * is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param characteristic The name of the characteristic
   * @param data       Data object containing the value returned.
   * @return           DeviceError reference, or null if no errors
   */
  //--------------------------------------------------------------//
  public DeviceError getCharacteristic (String property, String characteristic, Data data) 
       throws IOException {
    
    Object characValue;

    try {
      // getting the characteristicValue (or the corresponding deviceProperty)
      DeviceProperty devProperty= getDeviceClass().getDeviceProperty(property);
      characValue = devProperty.getCharacteristicValue(characteristic);
    }
    catch(Exception e) {
      // creating a deviceEvent with the appropriate DeviceError
      String dt_prop = new String(property);
      dt_prop = dt_prop.concat("_").concat(characteristic);
      DeviceEvent devEvent = new DeviceEvent (this,DeviceTransaction.GET,dt_prop,null, null);
      DeviceError devErr =  new DeviceError(devEvent,-1,e.getMessage());
      devEvent.error = devErr;
      return devErr;
    }
    //if returned object is a DataEntry, return the value in data
    if (characValue instanceof DataEntry) {
      if (data == null) 
	throw new IllegalArgumentException("Data object parameter can't be null");
      data.addDataEntry((DataEntry) characValue);
      return null;
    }
    // if returned object is a DeviceProperty, perform a get with this property
    else if (characValue instanceof DeviceProperty) {
      return this.get(((DeviceProperty)characValue).getName(), data);
    }
    throw new ClassCastException("no valid class returned by getCharacteristicValue");
  }
  
  /**
   * Initiate an asynchronous get value of a property's characteristic. An acknowledgement
   * is delivered to the listener when a response is received. If a 
   * response is not received within TIMEOUT seconds, an error is returned.
   * @param property   The name of the property
   * @param characteristic The name of the characteristic
   * @param listener   The device listener object to receive the response.
   */
  //--------------------------------------------------------------//
  public void getCharacteristic (String property, String characteristic, DeviceListener listener) 
       throws IOException {
    
    Object characValue;

    try {
      // getting the characteristicValue (or the corresponding deviceProperty)
      DeviceProperty devProperty= getDeviceClass().getDeviceProperty(property);
      characValue = devProperty.getCharacteristicValue(characteristic);
    }
    catch(Exception e) {
      // creating a deviceEvent with the appropriate DeviceError
      String dt_prop = new String(property);
      dt_prop = dt_prop.concat("_").concat(characteristic);
      DeviceEvent devEvent = new DeviceEvent (this,DeviceTransaction.GET,dt_prop,null, null);
      DeviceError devErr =  new DeviceError(devEvent,-1,e.getMessage());
      devEvent.error = devErr;
      listener.deviceChanged(devEvent);
      return;
    }
    //if returned object is a DataEntry, return the value in data in a DeviceEvent
    if (characValue instanceof DataEntry) { 
      Data d = new Data();
      d.addDataEntry((DataEntry) characValue);
      StringBuffer dt_prop= new StringBuffer(property);
      dt_prop= dt_prop.append("_").append(characteristic);
      DeviceEvent devEvent = new DeviceEvent (this,
					      DeviceTransaction.GET,
					      dt_prop.toString(),
					      d,
					      null);
      listener.deviceChanged(devEvent);
      return;
    }
    // if returned object is a DeviceProperty, perform a get with this property
    else if (characValue instanceof DeviceProperty) {
      this.get(((DeviceProperty)characValue).getName(), listener);
      return;
    }
    throw new ClassCastException("no valid class returned by getCharacteristicValue");
  }

  //-------------------------------------------------------------//
  // Accessors & mutators
  //-------------------------------------------------------------//
  /**
   * Return the name of the device.
   */
  public String getName ()    { return deviceName; }

  /**
   * Return the reference to the ControlSystem associated with the device.
   */
  public ControlSystem   getControlSystem ()  { return parentSys;  }

  /**
   * Return a reference to the context associated with the device.
   */
  public DeviceContext getContext () { return context;    }

  /**
   * Set the context for the device.
   * @param ctx    Data object containing context items
   */
  public void setContext (DeviceContext ctx) 
  {
    context = ctx;
    if (ctx == null) ctx = new DeviceContext();
  }

  /**
   * (Deprecated; use setContext (DeviceContext ctx) instead)
   * Set the context for the device.
   * @param data    Data object containing context items
   */
  public void setContext (Data data) 
  {
    context = new DeviceContext(data);
  }

  public Object getServiceData() { return serviceData;}
  public void setServiceData(Object sd) { serviceData = sd;}

  /*
   * returns the IO service bound to this device.
   */
  IOService getIOService () throws IOException {
    if (IOservice == null)
      try {
	IOservice = Config.getIOService(this);
      } catch (Exception e) {
	  throw new IOException ("Cannot obtain an I/O service for "
				  + this.getName()
				  + ": " + e);
      }
    return IOservice;
  }

  //-------------------------------------------------------------//
  // Directory service methods
  //-------------------------------------------------------------//

  private DeviceData deviceData;  // requires a directory service.

  /**
   * Returns the deviceData that describes this device.
   * @exception DataNotFoundException if this device is not defined 
   *            in the directory service or if no valid directory service
   */
  private DeviceData getDeviceData() throws DataNotFoundException {
    if (deviceData == null) {
      try {
	DirectoryService dir = Config.getDirectoryService();
	if (dir != null) deviceData = dir.getDeviceData(deviceName);
      } catch (Exception e) {
	  throw new DataNotFoundException ("Cannot get DeviceData for " 
					   + deviceName
					   + ": " + e);
      }
    }
    return deviceData;
  }

  /**
   * Returns the reference to the deviceClass this device belongs to.
   * @exception DataNotFoundException if no directory service, no device data
   *             or no device class defined.
   */
  public DeviceClass getDeviceClass() throws DataNotFoundException{    
    if (deviceData == null)
      getDeviceData ();
    if (deviceData != null) 
      return deviceData.getDeviceClass();
    else
      return null;
  }

  /**
   * Returns true if the I/O may require a cycle type.
   * @exception DataNotFoundException if no directory service or no device data
   */
    public boolean needCycleType() throws DataNotFoundException {
      if (deviceData == null)
	  getDeviceData ();
      if (deviceData != null)
	return deviceData.needCycleType();
      else
	return false;
    }

  /**
   * Returns true if the I/O with this property requires a cycle type.
   */
    public boolean needCycleType(String propertyName) throws DataNotFoundException {
      if (needCycleType() == false) 
	return false;			// this device never requires a cycle type.
      DeviceClass dc = getDeviceClass();
      if (dc != null) 
	return dc.getDeviceProperty(propertyName).needCycleType();
      else
	return false;
    }

}  // end of cdevDevice class
