//---------------------------------------------------------------------------
// Copyright (c) 1996,1997 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      This is a general purpose queue class that can queue up Objects.  It
//      is implemented as an array of objects.  It will automatically grow
//      the array when the occupancy count gets to the margin.  The margin
//      is set to a loadFactor of the capacity.
//
//  Author: Chip Watson
//
//  Based on work by Cuong Quach
//                                       
//----------------------------------------------------------------------------
// Revision History:
//

package cdev;

import java.io.*;
import java.util.*;

// **************************************************************************
/**
 * Fifo:
 * This is a general purpose fifo class that can deal with primitive int's.  
 * It is implemented as an array of int, which will automatically grow.
 */
// **************************************************************************

class Fifo
{
  public static final int DEFAULT_SIZE = 50;

  private Object[] ring = null;
  private int      nextRead = 0;  // indx to first filled data item
  private int      nextWrite = 0; // index to next available empty spot

  // Fifo: Constructors

  /**
   * Construct a Fifo with the given capacity.
   */

  public Fifo (int capacity)
  {
    ring = new Object[capacity];
  }

  /**
   * Construct a Fifo with the default capacity.
   */

  public Fifo ()
  {
    ring = new Object[DEFAULT_SIZE];
  }

  /**
   *  add:
   *  Place an int into the fifo.
   */
  public void add (Object obj)
  {
    synchronized (this) {              // protect agains multiple writer threads
      //----- append the object --------------//
      ring[nextWrite++] = obj;

      //----- check for index wrap ----------//
      if (nextWrite >= ring.length) nextWrite = 0;

      //----- check for full ring  -----------//
      if ( nextWrite == nextRead )  {        // ring is now full

	//---- copy the old data to the new array
	// this is done so that reader is oblivious to the growth
	Object [] newring = new Object[ring.length*2];
	System.arraycopy(ring,nextRead,newring,nextRead,ring.length-nextRead);
	// for (int i=nextRead; i<ring.length; i++) newring[i] = ring[i];
	// writer is one end-of-array wrap ahead of reader
	// reader will now miss a wrap, so copy low data to new 2nd half
	// and "undo" the last writer wrap
	System.arraycopy(ring,0,newring,ring.length,nextWrite);
	// for (int i=0; i<nextWrite; i++) newring[i+ring.length] = ring[i];
	nextWrite += ring.length;
	ring = newring;
	//System.out.println("ring size now "+ring.length);
	//System.out.println("reader "+nextRead +" writer "+nextWrite);
      }
    }
  }


  /**
   *  next:
   *  Return next int if available.  If no more 
   *  ints are in the fifo, then return -1.
   */
  public Object next ()
  {
    // may be able to lock less often if single reader assumed
    Object result;
    synchronized (this) { 

      //----- is fifo empty? ------------//
      if (nextRead == nextWrite) {
	result = null;
      
      //----- not empty  ----------------//
      } else {
	result = ring[nextRead++];

	//----- check for array index wrap -----//
	if ( nextRead >= ring.length) nextRead = 0;
      }
    }
    return result;
  }

//-------------------------------------------------------------//
// convenience functions
//-------------------------------------------------------------//

  /**
   * Returns the number of int's currently in the fifo.
   */
  public int size() {
    int len = ring.length; // protect against change
    int occupancy = nextWrite - nextRead;
    if (occupancy < 0) occupancy += len;
    return occupancy;  
  }

  /** 
   * Returns TRUE if the fifo is empty.
   */
  public boolean isEmpty ()  { return (nextRead==nextWrite)?true:false; }

  /** 
   * Returns TRUE if the fifo is not empty.
   */
  public boolean notEmpty ()  { return (nextRead!=nextWrite)?true:false; }

}
