//-----------------------------------------------------------------------------
// Copyright (c) 1996,1997 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//-----------------------------------------------------------------------------
//
//  Description:
//      Java implementation of cdevDataEntry class.
//
//  Issues:
//	For java arrays, which is the "first" dimension? 
//
//  Author: Chip Watson
//  Modified: Marc Constanty (M.C)
//  
//  From work by Danjin Wu and Walt Akers
//                                       
//-----------------------------------------------------------------------------
// Revision History:
//
//  2-10-97:  Remove cdev prefix from class name references.
//  1-15-98:  Modified for higher performance, JDK 1.1 (reflection)
//  1-28-98:  Added primitive constructors, array support;
//            replaced bounds with dimensions
//  2-02-98:  Removed support for the Java type TimeStamp (use double)
//  9-03-98:  FDM added support for String data.
// 10-20-98:  M.C: Cloning support added, functions clone(), and deepArrayClone() added
// 11-06-98:  M.C: new function DataEntry copy added: deepArrayCopy() and dataCopyInto() 
// 11-25-98:  calcDims() completely modified, dimensions max returned now
// 03-31-99:  M.C added convertors access to long and short
//  2-Jun-99: CW added constructor support for Number (JDBC convenience)
//
//-----------------------------------------------------------------------------

package cdev.data;

import java.math.*;
import java.lang.reflect.*;

public final class DataEntry implements Cloneable {

  int key;          // indentifying key
  short type;	    // enumerated type, -1 = unknown
  short ndims;	    // -1 = unknown; 0 = primitive type for original data
  int dims[];       // array of dimensions, null if ndims=0 or -1
  Object data;	    // String, x[], x[][], x[][][], ... x=double/float/int/.../String

  static final int MAXDIMENSIONS = 10;
  static DataConvertor [] convertors;

  static {
    convertors = new DataConvertor [Data.MAXTYPE+1];
    convertors[Data.BYTE] = new ByteConvertor();
    convertors[Data.SHORT] = new ShortConvertor();
    convertors[Data.INT] = new IntConvertor();
    convertors[Data.LONG] = new LongConvertor();
    convertors[Data.FLOAT] = new FloatConvertor();
    convertors[Data.DOUBLE] = new DoubleConvertor();
    convertors[Data.STRING] = new StringConvertor();
  }

  //========================= Constructors ===================================

  DataEntry() {}

  /**
   * Construct a DataEntry from an object.
   * Valid object types are only [multidimensional] arrays 
   * of either primitive numeric type or String.
   */
  public DataEntry(int newkey, Object d) {
    key = newkey;
    if (d instanceof String) {
      String [] temp = new String[1];
      temp[0] = (String) d;
      data = temp;
      ndims = 0;
      type = Data.STRING;
      return;
    } else if (d instanceof Number) {
      Number num = (Number) d;
      if (d instanceof Double || d instanceof BigDecimal) {
	double [] temp = new double[1];
	temp[0] = num.doubleValue();
	data = temp;
	type = Data.DOUBLE;
      } else if (d instanceof Float) {
	float [] temp = new float[1];
	temp[0] = num.floatValue();
	data = temp;
	type = Data.FLOAT;
      } else if (d instanceof Long || d instanceof BigInteger) {
	long [] temp = new long[1];
	temp[0] = num.longValue();
	data = temp;
	type = Data.LONG;
      } else if (d instanceof Integer) {
	int [] temp = new int[1];
	temp[0] = num.intValue();
	data = temp;
	type = Data.INT;
      } else if (d instanceof Short) {
	short [] temp = new short[1];
	temp[0] = num.shortValue();
	data = temp;
	type = Data.SHORT;
      } else if (d instanceof Byte) {
	byte [] temp = new byte[1];
	temp[0] = num.byteValue();
	data = temp;
	type = Data.BYTE;
      }	
      ndims = 0;
      return;
    }	  
    data = d;
    ndims = -1;		// unknown:
    type = -1;		// type and dimensions not calculated yet
  }

  /**
   * Construct a DataEntry from a double
   */
  public DataEntry(int newkey, double d) {
    key = newkey;
    double [] temp = new double[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.DOUBLE;
  }

  /**
   * Construct a DataEntry from a float
   */
  public DataEntry(int newkey, float d) {
    key = newkey;
    float [] temp = new float[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.FLOAT;
  }

  /**
   * Construct a DataEntry from a long
   */
  public DataEntry(int newkey, long d) {
    key = newkey;
    long [] temp = new long[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.LONG;
  }

  /**
   * Construct a DataEntry from a int
   */
  public DataEntry(int newkey, int d) {
    key = newkey;
    int [] temp = new int[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.INT;
  }

  /**
   * Construct a DataEntry from a short
   */
  public DataEntry(int newkey, short d) {
    key = newkey;
    short [] temp = new short[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.SHORT;
  }

  /**
   * Construct a DataEntry from a byte
   */
  public DataEntry(int newkey, byte d) {
    key = newkey;
    byte [] temp = new byte[1];
    temp[0] = d;
    data = temp;
    ndims = 0;
    type = Data.BYTE;
  }

  /**
   * Construct a DataEntry with specified key, dimensions, and type.
   * If array of dimensions is null, create a DataEntry with an array with one
   * element as data.
   * @param    newKey   The key of DataEntry
   * @param    newType   an int for the type
   * @param    newDims   an array of int for the array of dimensions
   * @exception   IllegalArgumentException if invalid type
   */

  // Author: M.C

  public DataEntry(int newKey, int newType, int[] newDims) throws IllegalArgumentException {

    key = newKey;
    type = (short) newType;
    if (type == Data.INVALID) throw new IllegalArgumentException("Invalid Type");
    if (newDims != null) {
      ndims = (short) newDims.length;
      dims = new int[ndims];
      System.arraycopy(newDims,0,dims,0,ndims);
      data = Array.newInstance(Data.localTypeToClass[type],dims);
      
    } else { //if dims == null, ndims is supposed to be 0
      ndims = 0;
      dims = null;
      data = Array.newInstance(Data.localTypeToClass[type],1);
    }
  }


  // ========================= Access functions ==============================

  /**
   * Get the array of dimensions. Returns null if data is scalar.
   */
  public int[] getDimensions() {
    if (ndims == 0) // scalar
      return null;
    if (ndims == -1)
      calcDims();
    return dims;
  }

  /**
   * Get the number of dimensions.
   */
  public int getDim() {
    if (ndims == -1)
      calcDims();
    return ndims;
  }

  /**
   * Get the integer key.
   */
  public int getKey() {
    return key;
  }

  /**
   * Get the key as a String.
   */
  public String getKeyAsString() {
    return Data.convertKey(key);
  }

  /**
   * Change the integer key, and return the old key.
   */
  public int setKey(int ikey) {
    int oldkey = key;
    key = ikey;
    return oldkey;
  }

  /**
   * Get the component data type (returns the Class).
   */
  public Class getComponentType ()
  {
    Class c = data.getClass();
    while (c.isArray()) 
      c = c.getComponentType();
    return c;
  }

  /**
   * Get the enumerated data type (returns an int type).
   */
  public int getEnumeratedType ()
  {
    if (type == -1)
      calcType();
    return type;
  }

  /**
   * Get the number of elements in the data array (if scalar, return 1).
   */
  public int nelems() {
    if (ndims==0) return 1;
    int ne = 1;
    if (ndims== -1)
      calcDims();
    for (int idim=0; idim<ndims; idim++) {
      ne *= dims[idim];
    }
    return ne;
  }


  /******************** Data Extraction Routines **************************/

  /**
   * Get the contained data.
   */
  public Object getValue() { return data; }

  /**
   * Get a double from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public double doubleValue() {
    if (type == -1) calcType();
    return convertors[type].doubleValue(data);
  }

  /**
   * Get a float from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public float floatValue() {
    if (type == -1) calcType();
    return convertors[type].floatValue(data);
  }

  /**
   * Get a long from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public long longValue() {
    if (type == -1) calcType();
    return convertors[type].longValue(data);
  }

  /**
   * Get an integer from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public int intValue() {
    if (type == -1) calcType();
    return convertors[type].intValue(data);
  }

  /**
   * Get a short from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public short shortValue() {
    if (type == -1) calcType();
    return convertors[type].shortValue(data);
  }

  /**
   * Get a byte from the DataEntry.  An Exception will be thrown if
   * the requested conversion is not possible.
   */
  public byte byteValue() {
    if (type == -1) calcType();
    return convertors[type].byteValue(data);
  }

  /**
   * Return the string value of data in this DataEntry object.
   */
  public String stringValue() {
    if (type == -1) calcType();
    return convertors[type].stringValue(data);
  }

  /**
   * Return the data as an array of doubles.
   */
  public double[] doubleArray() {
    if (type == -1) calcType();
    return convertors[type].doubleArray(data);
  }

  /**
   * Return the data as an array of floats.
   */
  public float[] floatArray() {
    if (type == -1) calcType();
    return convertors[type].floatArray(data);
  }

  /**
   * Return the data as an array of integers.
   */
  public int[] intArray() {
    if (data instanceof int[]) return (int [])data;
    if (type == -1) calcType();
    return convertors[type].intArray(data);
  }

  /**
   * Return the data as an array of shorts.
   */
  public short[] shortArray() {
    if (type == -1) calcType();
    return convertors[type].shortArray(data);
  }
 
  /**
   * Return the data as an array of longs.
   */
  public long[] longArray() {
    if (type == -1) calcType();
    return convertors[type].longArray(data);
  }

  /**
   * Return the data as an array of bytes.
   */
  public byte[] byteArray() {
    if (type == -1) calcType();
    return convertors[type].byteArray(data);
  }

  /**
   * Return the data as an array of Strings.
   */
  public String[] stringArray() {
    if (type == -1) calcType();
    return convertors[type].stringArray(data);
  }


  /**
   * Convert the entire DataEntry object to a string.  The resulting
   * string will contain tag, tag name, data type, and, when applicable,
   * dimensions.  If the data is scalar, it will be appended as well.
   * If data is 1D array, the first 8 elements will be appended.
   */
  public String toString() {
    if (type == -1) 
      calcType();
    if (ndims == -1)
      calcDims();
    StringBuffer result = new StringBuffer("Key:       ").append(key);
    result.append(" ").append(TagTable.convertKey(key)).append("\n");
    result.append("Data Type: "+ type + " ");
    result.append(data.getClass().getName()).append("\n");
    if (ndims != 0) {
      result.append("Dimensions: ");
      for (int i=0; i<dims.length; i++) {
	result.append("[").append(dims[i]).append("]");
      }
      result.append("\n");
    }
    if (ndims == 0) {
      result.append("Value:     ").append(stringValue()).append("\n");
    } else if (dims.length == 1) {
      result.append("Values:    ");
      int imax = dims[0];
      if (imax > 8) imax = 8;
      String [] strings = stringArray();
      for (int i=0;i<imax;i++) {
	result.append(strings[i]).append(" ");
      }
      result.append("...\n");
    }
    return result.toString();
  }

  // * * * * * * * * *  Tools Routines  * * * * * * * * * * * * * 

  /**
   * Creates a new object whose content is a full copy of this data entry.
   * Executes a deep copy of the contents of this data entry.
   * @exception CloneNotSupportedException if the value is not of one of the supported data types.
   */
  // Author: M.C
  public synchronized Object clone() throws CloneNotSupportedException,IllegalArgumentException{

    // we calculate dimensions and types
   
    if (ndims==-1) {calcDims();}
    if (type==-1)  {calcType();}

    // case of INVALID data
    if (type==Data.INVALID) {throw new CloneNotSupportedException("INVALID TYPE FOR DATA");
    }

    // cloning the DataEntry structure
    
    DataEntry d = (DataEntry) super.clone();
   
    // case of ndims==0 (primitive type)
    if (ndims==0) {
      d.dims = null;   // no array for dims
      Class c0= data.getClass();
      d.data= Array.newInstance(c0.getComponentType(),1);
      System.arraycopy(data,0,d.data,0,1); //copy of the one element array
    }

    // we are now cloning the array of dimensions in case ndims<>0
    else { 
      d.dims= new int[dims.length];
      System.arraycopy(dims,0,d.dims,0,dims.length);
    
      // we are now cloning the data array in case ndims<>0
      d.data = deepArrayClone(data);
    }
    // returning the new object
    return d;

  }


  /** 
   * Performs a compare & copy on the content of a data entry.
   * This data entry is compared with a specified destination data entry.
   * If this data entry and the destination one don't have the same structure:
   * same dimensions and same component type, the comparison fails and a clone
   * of this data entry is returned.
   * Otherwise, the contents of the two data entries are compared
   * and the contents of this data entry is copied into the destination one.
   * The keys are not compared.
   * @param dst  The destination data entry. If null, a clone of this data entry is returned.
   * @return  null if the comparison succeeds. 
   * Otherwise: either the destination data entry or a new clone of this data entry.
   * @see clone()
   */

  // Author: M.C

  public final synchronized DataEntry smartCompareCopy (DataEntry dst) throws CloneNotSupportedException {
  
    // if dst exists
    if (dst != null) {
      // get dimensions of source if not already known
      if (this.ndims == -1) this.calcDims();
      // get component type of source if not already known
      if (this.type == -1) this.calcType();
      // get dimensions of dst if not already known
      if (dst.ndims == -1) dst.calcDims();
      // get component type of dst if not already known
      if (dst.type == -1) dst.calcType();
      // compare dimensions and type, if matching compare data
      if ((this.type==dst.type) && (this.ndims==dst.ndims)) {
	int dimComp =0;
	//case of ndims<>0: previous test is enough if ndims==0
	if (ndims>0) {
	  for(int i=0;i<dims.length;i++) { // compare dims of src and dst
	    if (this.dims[i] != dst.dims[i]) {
	      dimComp =-1;
	      break;
	    }
	  }
	}
	if (dimComp==0) { // if dims are matching
	  // compare data and copy if necessary, clone if deepSmartArrayCopy doesn't work
	  boolean compResult = true;
	  try {
	    compResult = deepSmartArrayCopy(this.data,dst.data); //just perform a copy of data
	  }
	  catch (Exception e) { //if deepArrayCopy doesn't work
	    DataEntry entry = (DataEntry) this.clone(); //perform clone()
	    return(entry);
	  }
	  // if data fields are matching, no copy was done so return null
	  if (compResult) { 
	    return(null);
	  }
	  // if not,copy has been performed so return dst DataEntry
	  else {
	    return(dst);
	  }
	}
      }
    }  
    DataEntry entry = (DataEntry) this.clone();
    return(entry); // else, return a source clone.
  } 


  // * * * * * * * * * Private Routines * * * * * * * * * * * * *

  // Note: this routine could be further optimized for 1d arrays
  private void calcType() {
    if (data instanceof float[]) {
      type = Data.FLOAT;
      return;
    } else if (data instanceof int[]) {
      type = Data.INT;
      return;
    }
    Object o = data;
    Class c = o.getClass();
    while (c.isArray()) {
      c = c.getComponentType();
      if (c.isPrimitive())
	break;
      o = ((Object[])o)[0];
    }
    if (c==Byte.TYPE)
      type = Data.BYTE;
    else if (c==Short.TYPE)
      type = Data.SHORT;
    else if (c==Integer.TYPE)
      type = Data.INT;
    else if (c==Long.TYPE)
      type = Data.LONG;
    else if (c==Float.TYPE)
      type = Data.FLOAT;
    else if (c==Double.TYPE)
      type = Data.DOUBLE;
    else if (o instanceof String)
      type = Data.STRING;
    else
      type = Data.INVALID;
  }


  /**
   * calculate ndims and dims[] by taking for each level of depth the biggest
   * array length as dimension.
   * Convention: data array with one dimension and one element has ndims=0
   * and dims=null
   */

  private void calcDims() {
    
    // security for not using this function if dimensions already treated
    if (ndims!=-1) return;
    // optimization for most used arrays with one dimension 
    if (data instanceof float[]) { 
      if (((float[])data).length == 1) {
	ndims = 0;
	dims = null;
      } else {
      ndims =1;
      dims = new int[1];
      dims[0] = ((float[])data).length;
      return;
      }
    } else if (data instanceof int[]) {
      if (((int[])data).length == 1) {
	ndims = 0;
	dims = null;
      } else {
         ndims =1;
         dims = new int[1];
         dims[0] = ((int[])data).length;
         return;
      }
    }
    // general case
    int [] dimensions = new int [MAXDIMENSIONS];
    for (int i=0;i<MAXDIMENSIONS;i++) dimensions[i] = 0;
    ndims = (short) deepDimsMaxCalc(data, dimensions, 0);
    if ((ndims==1) && (dimensions[0]==1)) { //new convention
      ndims = 0;
      dims = null;
    } else {
      dims = new int[ndims];
      System.arraycopy(dimensions,0,dims,0,ndims);
    }
  }

  /**
   * perform a deep copy of an array of objects.
   */
  // Author: M.C

  private final Object deepArrayClone(Object src) {
    
    // No explicit test is done to verify that src is an array.
    Class c = src.getClass();
    Class cElem = c.getComponentType();  //get component type of src
    int srcLength = Array.getLength(src); // get length of src
    Object dst = Array.newInstance(cElem,srcLength);  //hold space for dst
    System.arraycopy(src,0,dst,0,srcLength);  //copy src elements into dst
    // go deeply if components of src are still arrays
    if (cElem.isArray()) {
      for (int i=0;i<srcLength;i++) {
        Array.set(dst,i,deepArrayClone(Array.get(src,i)));  
      }
    }
    return dst;
  }

  /**
   * performs a copy of a multidimensionnal array src to dst.
   * the destination Object is supposed to exist and to match src's type 
   * and dimension. If src and dst already have same values, doesn't copy,
   * else, just copy values in last recursion if needed.
   * @return true if src and dst have same values, else return false.
   */

  // Author: M.C


  private final boolean deepSmartArrayCopy(Object src, Object dst) {
    
    // No explicit test is done to verify that src and dst are arrays of same type and dimension.
    boolean compareTest = true; // to test if arrays are the same
    Class c = src.getClass();
    Class cElem = c.getComponentType();  //get component type of src
    int srcLength = Array.getLength(src); // get length of src
    // go deeply if components of src are still arrays
    if (cElem.isArray()) {
      for (int i=0;i<srcLength;i++) {
	compareTest &= deepSmartArrayCopy(Array.get(src,i),Array.get(dst,i));  
      } 
    }
    else { //if elements of src are from a primitive type or String
      compareTest = arrayCompare(src,dst,type,srcLength); //compare elements
      if (!compareTest) { // if differences
	System.arraycopy(src,0,dst,0,srcLength);  //copy src elements into dst
      }
    }
    return(compareTest);
  }

  /**
   * compares two unidimensional arrays having same type and same length
   * type can be a primitive type or String. In other case, false will be return.
   * @return true if same values, false if differences or if type is undefined
   */

  // Author: M.C

  private final boolean arrayCompare(Object arrayA, Object arrayB,int dataType, int length) {
    
    // no explicit test is done to verify if  A and B are unidimensional arrays.
    // no explicit test is done to verify matching between type and dimension of A and B.

    switch(dataType) { //look type of DataEntry
      
    case Data.INT: { // case of int
      int[] a = (int[]) arrayA;
      int[] b = (int[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false); // return false at first difference found
      }
      break;
    }
    case Data.FLOAT: { //case of float
      float[] a = (float[]) arrayA;
      float[] b = (float[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false);  // return false at first difference found
      }
      break;
    }
    case Data.SHORT: {  // case of short
      short[] a = (short[]) arrayA;
      short[] b = (short[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false); // return false at first difference found
      }
      break;
    }
    case Data.BYTE: { //case of byte
      byte[] a = (byte[]) arrayA;
      byte[] b = (byte[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false); // return false at first difference found
      }
      break;
    }
    case Data.LONG: { // case of long
      long[] a = (long[]) arrayA;
      long[] b = (long[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false);  // return false at first difference found
      }
      break;
    }
    case Data.DOUBLE: { // case of double
      double[] a = (double[]) arrayA;
      double[] b = (double[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false); // return false at first difference found
      }
      break;
    }
    case Data.STRING: { // case of string
      String[] a = (String[]) arrayA;
      String[] b = (String[]) arrayB;
      for(int i=0;i<length;i++) {
	if (a[i] != b[i]) return(false); // return false at first difference found
      }
      break;
    }
    case Data.INVALID: { // case invalid type
      return(false);  // return false
    }
    default: return(false); // in other cases (should not happen) return false
    }
    return(true); // if everything is matching, return true
  }

  /**
   * fill an array with the maximum of the dimension at each level of the
   * data array. Depth for the array must be uniform for each index.
   * return maximum depth seen during recursion
   */

  private int deepDimsMaxCalc(Object src, int[] dimsTab, int level) {

    // No explicit test is done to verify that src is an array.
    int result = -1; //to calculate depth
    Class c = src.getClass();
    Class cElem = c.getComponentType();  //get component type of src
    int srcLength = Array.getLength(src); // get length of src
    if (srcLength>dimsTab[level]) dimsTab[level] = srcLength;//replace in dimsTab if bigger
    if (cElem.isArray()) { //if components are still arrays
      for (int i=0;i<srcLength;i++) { //do the same job for each index on next level
	result = deepDimsMaxCalc(Array.get(src,i) , dimsTab, level+1);
      }
    } else { // if not array of array
      result = level+1;  // maximum depth obtained
    }
    return result;
  }

}


