//---------------------------------------------------------------------------
// Copyright (c) 1991,1992 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//----------------------------------------------------------------------------
//
//  Description:
//      Java implementation of xdrClass.  This XDR translator will convert
//      XDR data from an input stream into Java native types.  It will take
//      data from a buffer passed in by caller.
//      XDR types that are not native to Java are cast up to next bigger size.
//  
//  Issue:
//      1) Solaris long = int = 4 bytes.  read_long and read_u_long takes 
//         advantage of this.  However, this may not be possible on other OS.
//
//  Author: Cuong Quach (Patrick)
//
//  From work by Mr. Daniel Van Olst of SLAC
//                                       
//----------------------------------------------------------------------------
// Revision History:
//
//  9-3-96  : Created CQ.
//  9-20-96 : Added "finalize" method to close socket. CQ.
//  11-11-96: Make the xdr simple type writting methods final.
//   2-02-98: Remove socket support and related code
//   8-17-98: Replace deprecated String constructor in get_string. FDM.
// 26-Apr-99: CW Removed all "catch" and throw exceptions
//
//----------------------------------------------------------------------------

package cdev.data;

import java.io.*;
import java.util.*;

public class xdrReader
{
  public static int XDR_CHAR_SIZE = 4;


  //--------- Private data to implement socket stream behavior ---//
  private  DataInputStream  xdrIn       = null;


//-------------------------------------------------------------//
// xdrReader constructor
// Read from an XDR formated data buffer.
//-------------------------------------------------------------//
  public xdrReader (byte[] buf) 
  { 
    xdrIn = new DataInputStream( new ByteArrayInputStream(buf) );
  }

  // Read from an InputStream
  public xdrReader (InputStream in) {
    xdrIn = new DataInputStream(in);
  }

//-------------------------------------------------------------//
// get_byte:
//-------------------------------------------------------------//
  public final byte get_byte() throws IOException
  {
    byte result = xdrIn.readByte();
    return (result);
  }

//-------------------------------------------------------------//
// get_char:
//-------------------------------------------------------------//
  public final char get_char () throws IOException
  {
    char result = xdrIn.readChar();  // read first 2 bytes
    result = xdrIn.readChar();  // read last 2 bytes
    return (result);
  }


//-------------------------------------------------------------//
// get_u_char:
//-------------------------------------------------------------//
  public final byte get_u_char () throws IOException
  {
    byte[] b  = new byte[XDR_CHAR_SIZE];
    byte result;

    xdrIn.readFully (b, 0, XDR_CHAR_SIZE);
    result = b[3];
    return (result);    
  }

//-------------------------------------------------------------//
  public final short get_short() throws IOException
  {
    short result = (short) xdrIn.readInt();
    return (result);
  } 

//-------------------------------------------------------------//
// get_u_short:
//-------------------------------------------------------------//
  public final int get_u_short() throws IOException
  {
    int result = xdrIn.readInt();
    return (result);
  } 

//-------------------------------------------------------------//
// get_int:
//-------------------------------------------------------------//
  public final int get_int () throws IOException
  {
    int result = xdrIn.readInt();
    return (result);
  }

//-------------------------------------------------------------//
  public final long get_u_int () throws IOException
  {
    long result = 0;
    long b0,b1,b2,b3;

    byte[] b = get_bytes(4);
    b0 = (b[0]<0)?(256+(int)b[0]):(b[0]); 
    b1 = (b[1]<0)?(256+(int)b[1]):(b[1]); 
    b2 = (b[2]<0)?(256+(int)b[2]):(b[2]); 
    b3 = (b[3]<0)?(256+(int)b[3]):(b[3]); 
    result = (b0 * 16777216) + (b1 * 65536) + (b2 * 256) + b3;
    
    return (result);
  }


//-------------------------------------------------------------//
  public final long get_long() throws IOException
  {
    long result = xdrIn.readInt();   // Solaris long = int = 4 bytes
    return (result);
  } 


//-------------------------------------------------------------//
  public final long get_u_long() throws IOException
  {
    long result = (long) get_u_int();
    return (result);   // Solaris long = int = 4 bytes
  }


//-------------------------------------------------------------//
  public final float get_float() throws IOException
  {
    float result = xdrIn.readFloat();
    return (result);
  } 


//-------------------------------------------------------------//
  public final double get_double() throws IOException
  {
    double result;
    result = xdrIn.readDouble();
    return (result);
  } 


//-------------------------------------------------------------//
  public final byte[] get_bytes  ( ) throws IOException
  {
    return (get_opaque());
  }


//-------------------------------------------------------------//
  public final byte[] get_bytes  (int count ) throws IOException
  {
    byte[] b = new byte[count];
    xdrIn.readFully (b, 0, count);
    return (b);
  }


//-------------------------------------------------------------//
  public final int get_string_len () throws IOException
  {
    return (get_int());
  }


//-------------------------------------------------------------//
  public final String get_string () throws IOException
  {
    byte[] b = get_opaque();     // read the bytes
    String s = new String(b);
    return (s);    
  }


//-------------------------------------------------------------//
  public final String[] get_string (int count) throws IOException
  {
    String[] strArray = new String[count];
    byte [] buffer = new byte[256];
    for (int i=0; i<count; i++) {
      int slen = get_int();
      if (slen > buffer.length) buffer = new byte[slen+16];
      if (slen > 0) {
	xdrIn.readFully (buffer, 0, slen);  // read data bytes
	if ((slen % 4) > 0 )   {               // check for pading
	  int padLen = 4 - (slen % 4);
	  xdrIn.skipBytes(padLen);  // skip over the padding
	}
      }
      strArray[i] = new String(buffer,0,slen);
      //System.out.println ("got string["+i+"] slen "+slen+": "+strArray[i]);
    }
    return (strArray);
  }


//-------------------------------------------------------------//
  public final byte[] get_opaque () throws IOException
  {
    int slen = get_int();                    // get the length of byte array
    byte[] result  = new byte[slen];    

    //System.out.println ("get_opaque: getting "+slen+" bytes");
    if (slen > 0)
      xdrIn.readFully (result, 0, slen);  // read data bytes
    if ((slen % 4) > 0 )   {               // check for pading
      int padLen = 4 - (slen % 4);
      xdrIn.skipBytes(padLen);  // skip over the padding
    }
    return (result);  
  }

}  // end of xdrReader class

