package cdev.dir;

import java.io.*;
import java.util.*;
import java.sql.*;

/**
 * Interface DeviceMessage for messages involving a device.
 * <P>
 * A DeviceMessage object is created by
 * <a href="cdev.dir.DirectoryService.html#getDeviceMessage(int,int)">
 * DirectoryService#getDeviceMessage(messageLevel, messageNumber)</a>
 * in the DirectoryService class.
 * <P>
 * A DeviceMessage can hold information about the state of a device, for
 * display by the program or for sending to related programs or alarm
 * collectors.
 * A message can contain all or part of the following information:
   <UL>
 * <LI><B>messageLevel</B> : a small integer indicating the nature and
 *     priority of the message. Some conventions are needed here.
 * <LI><B>messageNumber</B> : an integer identifying the message for the
 *     level.
 * <LI><B>category</B> : identifies general message categories such as
 *     COMMUNICATIONERROR. Some conventions are needed here.
 * <LI><B>label</B> : one-word name for the message.
 * <LI><B>shortMessage</B> : a one-line description.
 * <LI><B>text</B> : multiline text string with more complete information.
 *     If this information exists at all, it will be in the form of
 *     a reference to the database.
 * </UL>
 * The superclass DirectoryItem defines methods <I>getName</I> that returns
 * the symbolic name of the message and <I>getDescription</I> that returns
 * shortMessage.
 * <P>
 * General documentation in <a href="DirectoryService.html">DirectoryService</a> .
 *
 * @author   J.Cuperus  started Sep 1998;
 * @version  1.0
 *
 */
public interface DeviceMessage extends DirectoryItem {

//---------------------------------------------------------------------------

/**
 * Gets the one-word message label.
 *
 * @return one-word message label or else the string: "-"
 */
public String getLabel();

//---------------------------------------------------------------------------

/**
 * Gets the message category.
 * This indicates the meaning of some messages. Some conventions on
 * standard categories (like "COMMUNICATIONERROR") are needed.
 *
 * @return message category or else the string: "-"
 */
public String getCategory();

//---------------------------------------------------------------------------

/**
 * Gets the message level, indicating nature and priority of the message.
 *
 * @return the message level.
 */
public int getLevel();

//---------------------------------------------------------------------------

/**
 * Gets the number that identifies the message for the level.
 *
 * @return the message number.
 */
public int getNumber();

//---------------------------------------------------------------------------

/**
 * Gets the one-line message string.
 *
 * @return the one-line message string or else the string: "-".
 */
public String getShortMessage();

//---------------------------------------------------------------------------

/**
 * Gets the full text of the message, if any, from the database.
 *
 * @return the multiline message string.
 * @exception NoSuchDataException a multiline message is not available
 */
public String getLongMessage() throws NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * True when message text is available in the database.
 * This is a test to avoid the exception in getLongMessage.
 *
 * @return true if message text available
 */
public boolean isLongMessage();

//---------------------------------------------------------------------------

}
