package cdev.dir;

import java.io.*;
import java.util.*;
import java.sql.*;
import cdev.data.*;

/**
 * DeviceProperty accesses data associated with a property.
 * A property corresponds to a device value we can always read and
 * often also write.
 * <P>
 * Get a DeviceProperty object for an instance of DeviceClass with
 * <a href="cdev.dir.DeviceClass#getDeviceProperty(java.lang.String)">
 *    DeviceClass#getProperty(propertyName)</a>.
 * <P>
 * Get a list of attributes with
 * <a href="#getCharacteristicNames">
 *    getCharacteristicNames()</a> and a corresponding value with
 * <a href="#getCharacteristicValue">
 *    getCharacteristicValue(characteristicName)</a>.
 * <P>
 * You can also obtain a DeviceProperty object with
 * <a href="cdev.dir.DirectoryService.html#getDeviceProperty">
 * DirectoryService#getDeviceProperty(propertyName,className)</a>.
 * <BR>You can obtain a DeviceProperty object common to several classes with
 * <a href="cdev.dir.DirectoryService.html#getCommonProperty">
 * DirectoryService#getCommonProperty(propertyName,classNames)</a>.
 * <P>
 * This description should be completed with a set of conventions for
 * standard characteristics and their value domain.
 * <P>
 * General documentation in <a href="DirectoryService.html">DirectoryService</a> .
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DeviceClass
 * @see BitPatternDefinition
 * @see DeviceList
 */
public interface DeviceProperty extends DirectoryItem {

//---------------------------------------------------------------------------

/**
 * Return array of names of characteristics for the property.
 *
 * @return array of names of characteristics
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 */

public String[] getCharacteristicNames()
       throws NoSuchDataException, BadConnectionException;

//---------------------------------------------------------------------------

/**
 * Gets the value of the characteristic.
 * This is either a DataEntry object or else a DeviceProperty object when
 * the value must be obtained with a property.
 * The key for the DataEntry object is set to 0.
 *
 * @param characteristicName  the name of the characteristic
 * @return characteristic value as DataEntry or DeviceProperty object.
 * @exception NoSuchDataException    -
 * @exception BadConnectionException -
 * @exception BadQueryException      -
 */

public Object getCharacteristicValue(String characteristicName)
       throws NoSuchDataException, BadConnectionException, BadQueryException;

//---------------------------------------------------------------------------

/**
 * Gets the name of the device class to which the property belongs.
 *
 * @return the name of the device class
 */
public String getClassName();

//---------------------------------------------------------------------------

/**
 * Gets the number of dimensions.
 *
 * @return number of dimensions (0 if scalar)
 */

public int getDim();

//---------------------------------------------------------------------------

/**
 * Gets the array of dimensions.
 *
 * @return array of dimensions
 */

public int[] getDimensions();

//---------------------------------------------------------------------------

/**
 * Gets the enumerated data type of the property (returns an int).
 *
 * @return data type (enumeration)
 */

public int getEnumeratedType();

//---------------------------------------------------------------------------

/**
 * Return true if property value can be considered constant
 * and needs to be fetched only once (min, max, ...).
 *
 * @return true if property refers to a constant value
 */
public boolean isConstant();

//---------------------------------------------------------------------------

/**
 * Return true if the characteristic is defined for the property.
 *
 * @param characteristic  the name of the characteristic
 * @return true if the characteristic is defined for the property
 */

public boolean isDefined(String characteristic);

//---------------------------------------------------------------------------

/**
 * Returns true if this is a discrete property.
 * For discrete properties you can get further information with
 * <a href="#getDiscretePattern">getDiscretePattern</a>.
 *
 * @return true if this is a discrete property
 */

public boolean isDiscrete();

//---------------------------------------------------------------------------

/**
 * Return true if property can write values.
 *
 * @return true if property can write values
 */
public boolean isWritable();

//---------------------------------------------------------------------------

/**
 * Return true if property value depends on cycle type.
 *
 * @return true if property value depends on cycle type
 */
public boolean needCycleType();

//---------------------------------------------------------------------------

}
