package cdev.dir;

/**
 * The parent interface of all directory interfaces.
 * It defines some common functionality.
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DeviceData
 * @see DeviceClass
 * @see DeviceList
 * @see DeviceProperty
 * @see BitPatternDefinition
 * @see DeviceMessage
 */
public interface DirectoryItem {

//===========================================================================
// Instance Methods
//===========================================================================

/**
 * Gets a one-line description of the directory object.
 * Mainly documentation for display to users but, in some cases like in
 * ErrorData it can convey essential information.
 *
 * @return    the description of the directory object (null if none)
 */
public String getDescription();

//---------------------------------------------------------------------------

/**
 * Gets the name of the service which supplied the information.
 * In a mixed system, this identifies the subsystem that provides the
 * data. This information may also be used by the equipment access
 * classes to select the I/O implementation.
 *
 * @return    the name of the service that supplied the information
 */
public String getServiceName();

//---------------------------------------------------------------------------

/**
 * Abstract method: the implementation returns a multiline string with
 * information about the object.
 * If 'all' is true, more complete information may be returned.
 * To be implemented in the subclass.
 *
 * @param all true if you want complete information
 * @return    the multiline information string
 */
public String getInfo(boolean all);

//---------------------------------------------------------------------------

/**
 * Gets minimum number of lines to display the getInfo string.
 *
 * @return    the number of lines in the information string.
 */
public int getInfoLines();

//---------------------------------------------------------------------------

/**
 * Gets minimum line width, in characters, to display the getInfo string.
 *
 * @return    the number of characters per line
 */
public int getInfoWidth();

//---------------------------------------------------------------------------

/**
 * Gets the name of the directory object.
 *
 * @return    the name of the directory object
 */
public String getName();

//---------------------------------------------------------------------------

/**
 * Gets the identifier as a representation of the directory object.
 * This may be the name, if it is unique for the class, or a concatenation
 * of names like 'className/propertyName' which is unique.
 *
 * @return    the representation string for the directory object
 */
public String toString();

//---------------------------------------------------------------------------

}
