package cdev.dir;

import java.sql.*;

/**
 * This abstract class is a starting point for getting DirectoryItem
 * objects and object identifiers from string or numeric parameters.
 * After this you can get additional information from the objects
 * themselves.
 * <P>
 * As a first step you must get the implemented directory service
 * with a statement like:
 * <P>
 * DirectoryService dir = cdev.dir.service.DirectoryServiceImplementation.getService();
 * <P>
 * This will also connect you to any database(s) you
 * may need for getting data.
 * After this you work only with the interfaces with statements like:
 * <P>
 * DeviceData devdat = dir.getDeviceData("BR1.QNO");  and :
 * String devname = devdat.getName();
 * <P>
 * Several methods require a 'query' string for making a selection of
 * object identifiers. The form and content of these queries is
 * implementation dependent. For an implementation with a relational
 * database this could be any valid 'WHERE' section of an SQL query, like
 * 'column1=data1 AND column2=data2'. Avoid such queries in generic
 * programs which must be ported to other implementations.
 * <P>
 * General documentation in <a href="DirectoryServiceDoc.html">DirectoryServiceDoc</a> .
 *
 * Note: only one DirectoryServiceImplementation is installed for a site.
 *       This implementation may however address <a href="SubServicesDoc.html"> multiple DirectoryItem implementations</a>
 *
 * @author   J.Cuperus
 * @version  1.0
 *
 * @see DeviceData
 * @see DeviceClass
 * @see DeviceList
 * @see DeviceProperty
 * @see DeviceMessage
 * @see BitPatternDefinition
 * @see cdev.dir.service.DirectoryServiceImplementation
 */
public abstract class DirectoryService {

//===========================================================================
// Static Methods
//===========================================================================

/**
 * Gets unique instantiation of the directory service implementation.
 * This static method must be overridden in the implementation.
 *
 * @exception BadConnectionException -
 */
public static DirectoryService getService()
       throws BadConnectionException {
  return null;
};

//===========================================================================
// Constructor
//===========================================================================

/**
 * Default constructor, never called directly.
 */
protected DirectoryService() { }


//===========================================================================
// DeviceClass
//===========================================================================

/**
 * Gets DeviceClass object when given the class name.
 *
 * @param className  the name of the device class
 * @return the DeviceClass object
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract DeviceClass getDeviceClass(String className)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Gets an array of all device class names.
 *
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 */
public abstract String[] getDeviceClassNames()
       throws BadConnectionException, NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * Gets an array of device class names from a query string.
 *
 * @param query  a selection statement depending on the implementation
 * @return array of device class names
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception BadQueryException      -
 */
public abstract String[] getDeviceClassNames(String query)
       throws BadConnectionException, NoSuchDataException, BadQueryException;


//===========================================================================
// DeviceData
//===========================================================================

/**
 * Gets the DeviceData object from the device name.
 *
 * @return the DeviceData object.
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract DeviceData getDeviceData(String deviceName)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Gets device names from a query string.
 *
 * @param query  a selection statement depending on the implementation
 * @return array of device names
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception BadQueryException      -
 */
public abstract String[] getDeviceNames(String query)
       throws BadConnectionException, NoSuchDataException, BadQueryException;


//===========================================================================
// DeviceList
//===========================================================================

/**
 * Gets array of all devicelist names.
 *
 * @return array of device list names
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 */
public abstract String[] getDeviceListNames()
       throws BadConnectionException, NoSuchDataException;

//---------------------------------------------------------------------------

/**
 * Gets array of devicelist names from a database query.
 *
 * @return array of device list names
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception BadQueryException      -
 */
public abstract String[] getDeviceListNames(String query)
       throws BadConnectionException, NoSuchDataException, BadQueryException;

//---------------------------------------------------------------------------

/**
 * Gets DeviceList object from the name of a list.
 *
 * @param listName   the name of the device list
 * @return DeviceList object
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract DeviceList getDeviceListByName(String listName)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Gets DeviceList object from a database query string.
 * Devices are grouped by class. Ordering of groups and devices is
 * implementation-dependent.
 *
 * @param query  a selection statement depending on the implementation
 * @return array of device list names
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception BadQueryException      -
 */
public abstract DeviceList getDeviceListByQuery(String query)
       throws BadConnectionException, NoSuchDataException, BadQueryException;


//===========================================================================
// DeviceProperty
//===========================================================================

/**
 * Gets DeviceProperty object for given property name and class name.
 *
 * @param  className the name of the class
 * @param  propertyName the property name for the class
 * @return DeviceProperty object
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract DeviceProperty getDeviceProperty(String propertyName,
						 String className)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

/**
 * Instantiates common DeviceProperty object for property name and device class names.
 * This determines what can be done collectively with a set of devices
 * belonging to different classes, for instance a group from a DeviceList.
 * <P>
 * To be compatible, the properties must:
 * <UL>
 * <LI> have the same datatype
 * <LI> be all scalar or all array
 * <LI> have the same service name
 * </UL>
 * Throw NoSuchDataException if the classes are incompatible for the property.
 * <P>
 * The values of the composite property for the classes are :
 * <UL>
 * <LI> dataSize is the largest size
 * <LI> constantProp is false if one of them is false
 * <LI> writeProp is false if one of them is false
 * <LI> needCycleType is true if one of them is true
 * <LI> description is the first non null description
 * </UL>
 * <P>
 * Only the compatible characteristics defined in all the classes are listed.
 * To be compatible, characteristics which have a property name as value
 * must have the same name for all the classes. Characteristics which have
 * different values for some of the classes are given the value string
 * "MULTIVALUED"
 *
 * @param propertyName  the name of the property
 * @param classNames the classes for which the property should be defined
 * @return DeviceProperty object
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 */
public abstract DeviceProperty getCommonProperty(String propertyName,
					String[] classNames)
       throws BadConnectionException, NoSuchDataException;


//===========================================================================
// DeviceMessage
//===========================================================================

/**
 * Gets DeviceMessage object corresponding to level, number and service.
 * If the serviceName is irrelevant, set it to "ALL".
 *
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException - when duplicate messageKey found.
 */
public abstract DeviceMessage getDeviceMessage(int messageLevel,
				      int messageNumber, String serviceName)
  throws BadConnectionException, NoSuchDataException, TooManyValuesException;


//===========================================================================
// BitPatternDefinition
//===========================================================================

/**
 * Gets BitPatternDefinition object for the given pattern name.
 *
 * @param  patternName the pattern name for the class
 * @return BitPatternDefinition object
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract BitPatternDefinition getBitPatternDefinition(
						    String patternName)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;


//===========================================================================
// Symbols
//===========================================================================

/**
 * Gets the symbol value when given the symbol category and symbol name.
 *
 * @param symbolCategory the category of the symbol
 * @param symbolName     the name of the symbol
 * @return the corresponding symbol value
 * @exception BadConnectionException -
 * @exception NoSuchDataException    -
 * @exception TooManyValuesException -
 */
public abstract int getSymbolValue(String symbolCategory, String symbolName)
       throws BadConnectionException, NoSuchDataException, TooManyValuesException;

//---------------------------------------------------------------------------

}
